﻿within ;
package RandomLib "Library for generating random numbers and random variates"
  package Variates "Random variates generation package"
    class UsersGuide "Users Guide"
      extends Modelica.Icons.Information;

        annotation (Documentation(info="<html>
<h3><font color=\"#008000\" size=5>Users Guide of the Variates Library</font></h3>
 
<p>
<b>Variates</b> package contains a set of functions for generating  random observations from the following probability distributions.
</p>
Discrete Probability Distributions:
<ul>
<li> Empirical Discrete
<li> Bernoulli
<li> Discrete Uniform
<li> Binomial
<li> Geometric
<li> Negative Binomial
<li> Poisson
</ul>
 
Continuous Probability Distributions:
<ul>
<li> Empirical Continuous
<li> Uniform
<li> Exponential
<li> Erlang
<li> Gamma
<li> Weibull
<li> Normal
<li> LogNormal
<li> Beta
<li> Johnson (bounded and unbounded)
<li> Triangular
</ul>
 
<p>
Most of the algorithms used for generating the random variates are detailed in \"Simulation Modeling and Analysis\" (Averill M. Law, McGraw Hill, 2007), or the documentation of the Arena Simulation environment.<br>
</p>
 
 
<h3><font color=\"#008000\">Library Structure</font></h3>
<p>
The structure of the Variates package is divided in three components: 1) general type and functions; 2) Discrete package; and 3) Continuous package.<br><br>
The general type and functions mainly contain the <em>Generator</em> record, which represent the source for uniform random numbers that will be used by 
the <em>U01</em> function, which is the function that generates uniform random numbers.<br>
Each <em>Generator</em> must be initialized using the function <em>initGenerator</em>.<br>
<em>Var</em> function is a prototype for the rest of variate generation functions.<br>
<em>GenerateVariate</em> is a function that accepts an integer parameter that indicates the probability distribution for the random variate.<br>
The <em>Constant</em> function generates a constant number, indicated as a parameter.<br>
<br>
 
Package <em>Discrete</em> contains the functions for generating random variates following discrete probability distributions.<br><br>
 
Package <em>Continuous</em>, analogously to the previous, contains the functions for generating random variates following continuous probability distributions.<br><br>
 
Each function of these packages contains a description of the random variates that generates.
</p>
 
<p>
 
</p>
 
<h3><font color=\"#008000\">Random Variate Generation</font></h3>
<p> 
In order to generate uniform random observations using Variates package the user has to:
<ul>
<li> Declare a Generator that represents the source of uniform random numbers.
<li> Initialize it with the initGenerator function.
<li> Call the function of the desired probability distribution with the required parameters to generate a random variate. This function will return the generated random variate and the updated state of the Generator.
</ul>
 
Notice that the updated state of the Generator has to be used to generate the next random number.<br>
</p>
<p>
The following example represents the steps indicated above.
 
<pre>
 model usingVariates 
  // declaration of the generator 
  Variates.Generator g1;
  Real u1[6];
algorithm 
  when initial() then
    // generator initialization  
    g1 := Variates.initGenerator();
  end when;
  when time &lt;= 0 then
    for i in 1:6 loop
      // generation of random variates from g1 with Exponential(8) distribution.
      (u1[i],g1):= Variates.Continuous.Exponential(g1,8);
    end for;
  end when;
end usingVariates;
 
//results: u1 = {1.20419,12.282,0.713961,5.28482,7.26866,41.0291}
 
</pre>
</p>
 
<p>
Several independent sources of random numbers can be created by declaring several Generators (RngStreams). The initialization of each declared Generator will give it a different initial state.<br>
The package seed is automatically managed, so it is not required to set an initial seed.
 
</p> 
 
<h3><font color=\"#008000\">Use Another Source of Uniform Random Numbers</font></h3>
<p>
CMRG is the default uniform random number generator for the Variates package. However, Variates package can be used with any other Modelica library for uniform random generation. 
<br>
To use other uniform random number generator with the Variates package the following requirements have to me meet:
<ul>
<li> The random number generator has to be described by a Modelica record datatype. For example, for a generator that reads numbers from a file, this record will store the filename and the last line read.
<li> The record that describe the random number generator has to be initialized using a function. This function can not have input parameters.
<li> A function to generate random numbers  has to be declared following this prototype:
<pre>
  function name
    input generator g;
    output Real u;
    output generator gout;
  end name;
</pre>
Where generator is the record mentioned in the previous requirement.
</ul><br>
Once the new random number generator meets these requirements, the Variates Generator, initGenerator and U01 values has to be set to the ones in the new random number generator.
</html>"));
    end UsersGuide;

    record Generator = CMRG.RngStream
      "Uniform random number source (generator)";
    function initGenerator = CMRG.CreateStream
      "Uniform random source initialization function"
    annotation (preferredView="info",Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
initGenerator
</font></h3>
<p>
Initializes the random number generator
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>initGenerator</b>()</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Reads the value of the package seed, initializes the Generator and updates the package seed. 
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
g := Variates.initGenerator();
</pre>
 
</HTML>
"));
    function initGeneratorWrap
      "initializes a RNG generator, returning its single components"
      output Real Cg1;
      output Real Cg2;
      output Real Cg3;
      output Real Cg4;
      output Real Cg5;
      output Real Cg6;
      output Real Bg1;
      output Real Bg2;
      output Real Bg3;
      output Real Bg4;
      output Real Bg5;
      output Real Bg6;
      output Real Ig1;
      output Real Ig2;
      output Real Ig3;
      output Real Ig4;
      output Real Ig5;
      output Real Ig6;
      output Integer Anti;
      output Integer IncPrec;
    protected
      Generator g;
    algorithm
      g := initGenerator();
      Cg1 := g.Cg[1];
      Cg2 := g.Cg[2];
      Cg3 := g.Cg[3];
      Cg4 := g.Cg[4];
      Cg5 := g.Cg[5];
      Cg6 := g.Cg[6];
      Bg1 := g.Bg[1];
      Bg2 := g.Bg[2];
      Bg3 := g.Bg[3];
      Bg4 := g.Bg[4];
      Bg5 := g.Bg[5];
      Bg6 := g.Bg[6];
      Ig1 := g.Ig[1];
      Ig2 := g.Ig[2];
      Ig3 := g.Ig[3];
      Ig4 := g.Ig[4];
      Ig5 := g.Ig[5];
      Ig6 := g.Ig[6];
      Anti := g.Anti;
      IncPrec := g.IncPrec;
    annotation (preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
initGeneratorWrap
</font></h3>
<p>
Initializes the random number generator (returning its components separately)
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>initGeneratorWrap</b>()</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Reads the value of the package seed, initializes the Generator and updates the package seed. <br>
This function can be used if the components of the CMRG generator are declared using independent variables, instead of a record.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(g.Cg1,g.Cg2,g.Cg3,g.Cg4,g.Cg5,g.Cg6,<br>
g.Bg1,g.Bg2,g.Bg3,g.Bg4,g.Bg5,g.Bg6<br>
g.Ig1,g.Ig2,g.Ig3,g.Ig4,g.Ig5,g.Ig6<br>
g.Anti,g.IncPrec) := Variates.initGeneratorWrap();
</pre>
 
</HTML>
"));
    end initGeneratorWrap;

    function U01 = CMRG.RandU01 "Uniform random number generation function"
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
U01
</font></h3>
<p>
Uniform random number generation function
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>U01</b>(g)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Returns a uniform random number in the interval (0,1), and updates the state of the Generator g by one step
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
model Example 
  CMRG.RngStream g;
  Real u[5];
algorithm 
  when initial() then
    g := Variates.initGenerator();
  end when;
  when time &lt;= 0 then
    for i in 1:5 loop
      (u[i],g) := Variates.U01(g);
    end for;
  end when;
end Example;
 
// u = {0.988831,0.760616,0.855857,0.546418,0.868702}
 
</pre>
 
</HTML>
"));
    partial function Var "prototype for variate generation functions"
      input Generator g "Random number generator";
      input Real a = 0 "First parameter of the prob. distribution";
      input Real b = 0 "Second parameter of the prob. distribution";
      input Real c = 0
        "Third parameter of the prob. distribution (only for the Johnson and Triangular distributions)";
      input Real d = 0
        "Fourth parameter of the prob. distribution (only for the Johnson distribution)";
      output Real x "Generated random variate";
      output Generator gout "Updated random number generator";
    protected
      Real u;
    end Var;

    function GenerateVariate "Generates Variates"
      input Integer distribution;
      input Generator g;
      input Real p1 = 0;
      input Real p2 = 0;
      input Real p3 = 0;
      input Real p4 = 0;
      output Real u;
      output Generator gout;
    algorithm
      if distribution == 1 then
        // constant
        (u,gout) := Constant(
              g,
              p1);
      elseif distribution == 2 then
        // bernoulli
        (u,gout) := Discrete.Bernoulli(
              g,
              p1);
      elseif distribution == 3 then
        // DU
        (u,gout) := Discrete.DiscreteUniform(
              g,
              p1,
              p2);
      elseif distribution == 4 then
        // Bin
        (u,gout) := Discrete.Binomial(
              g,
              p1,
              p2);
      elseif distribution == 5 then
        // Geom
        (u,gout) := Discrete.Geometric(
              g,
              p1);
      elseif distribution == 6 then
        // NegBin
        (u,gout) := Discrete.NegativeBinomial(
              g,
              p1,
              p2);
      elseif distribution == 7 then
        // Poisson
        (u,gout) := Discrete.Poisson(
              g,
              p1);
      elseif distribution == 8 then
        // U
        (u,gout) := Continuous.Uniform(
              g,
              p1,
              p2);
      elseif distribution == 9 then
        // Exp
        (u,gout) := Continuous.Exponential(
              g,
              p1);
      elseif distribution == 10 then
        // Erlang
        (u,gout) := Continuous.Erlang(
              g,
              p1,
              p2);
      elseif distribution == 11 then
        // Gamma
        (u,gout) := Continuous.Gamma(
              g,
              p1,
              p2);
      elseif distribution == 12 then
        // Weibull
        (u,gout) := Continuous.Weibull(
              g,
              p1,
              p2);
      elseif distribution == 13 then
        // N
        (u,gout) := Continuous.Normal(
              g,
              p1,
              p2);
      elseif distribution == 14 then
        // LN
        (u,gout) := Continuous.LogNormal(
              g,
              p1,
              p2);
      elseif distribution == 15 then
        // Beta
        (u,gout) := Continuous.Beta(
              g,
              p1,
              p2);
      elseif distribution == 16 then
        // Johnson
        (u,gout) := Continuous.Johnson(
              g,
              p1,
              p2,
              p3,p4);
      elseif distribution == 17 then
        // Triang
        (u,gout) := Continuous.Triangular(
              g,
              p1,
              p2,
              p3);
      else
        u := 0;
        gout := initGenerator();
      end if;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
GenerateVariate
</font></h3>
<p>
 
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>GenerateVariate</b>(distribution,g,p1,p2,p3,p4)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
This function generates a random variate from the distribution indicated with the integer pararameter \"distribution\", using the following value correspondence:
<ul>
<li> 1 = Constant
<li> 2 = Bernoulli
<li> 3 = Discrete Uniform
<li> 4 = Binomial
<li> 5 = Geometric
<li> 6 = Negative Binomial
<li> 7 = Poisson
<li> 8 = Uniform
<li> 9 = Exponential
<li> 10 = Erlang
<li> 11 = Gamma
<li> 12 = Weibull
<li> 13 = Normal
<li> 14 = LogNormal
<li> 15 = Beta
<li> 16 = Johnson (bounded and unbounded)
<li> 17 = Triangular
<li> other value = initalization of the generator (call to initGenerator() function).
</ul>
 
Using this function, the generation of random variates in models can be easily parametrized. 
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 (u,g) := GenerateVariate(2,g,0.4);
 // Generates a discrete variate (u) with Bernoulli(0.4) probability distribution.
 
 (u,g) := GenerateVariate(11,g,0.4,0.6);
 // Generates a continuous variate (u) with Gamma(0.4,0.6) probability distribution.
 
The example <a href=\"Modelica://DESLib.RandomLib.Examples.VariatesSimple2\">VariatesSimple2</a> uses of this function.
 
 
 
</pre>
 
</HTML>
"));
    end GenerateVariate;

    function GenerateVariateWrap "Generates Variates"
      input Real inCg1;
      input Real inCg2;
      input Real inCg3;
      input Real inCg4;
      input Real inCg5;
      input Real inCg6;
      input Real inBg1;
      input Real inBg2;
      input Real inBg3;
      input Real inBg4;
      input Real inBg5;
      input Real inBg6;
      input Real inIg1;
      input Real inIg2;
      input Real inIg3;
      input Real inIg4;
      input Real inIg5;
      input Real inIg6;
      input Integer inAnti;
      input Integer inIncPrec;
      input Integer distribution;
      input Real p1 = 0;
      input Real p2 = 0;
      input Real p3 = 0;
      input Real p4 = 0;
      output Real u;
      output Real Cg1;
      output Real Cg2;
      output Real Cg3;
      output Real Cg4;
      output Real Cg5;
      output Real Cg6;
      output Real Bg1;
      output Real Bg2;
      output Real Bg3;
      output Real Bg4;
      output Real Bg5;
      output Real Bg6;
      output Real Ig1;
      output Real Ig2;
      output Real Ig3;
      output Real Ig4;
      output Real Ig5;
      output Real Ig6;
      output Integer Anti;
      output Integer IncPrec;
    protected
      Generator g;
      Generator gout;
    algorithm
      g.Cg[1] := inCg1;
      g.Cg[2] := inCg2;
      g.Cg[3] := inCg3;
      g.Cg[4] := inCg4;
      g.Cg[5] := inCg5;
      g.Cg[6] := inCg6;
      g.Bg[1] := inBg1;
      g.Bg[2] := inBg2;
      g.Bg[3] := inBg3;
      g.Bg[4] := inBg4;
      g.Bg[5] := inBg5;
      g.Bg[6] := inBg6;
      g.Ig[1] := inIg1;
      g.Ig[2] := inIg2;
      g.Ig[3] := inIg3;
      g.Ig[4] := inIg4;
      g.Ig[5] := inIg5;
      g.Ig[6] := inIg6;
      g.Anti := inAnti;
      g.IncPrec := inIncPrec;
      if distribution == 1 then
        // constant
        (u,gout) := Constant(
              g,
              p1);
      elseif distribution == 2 then
        // bernoulli
        (u,gout) := Discrete.Bernoulli(
              g,
              p1);
      elseif distribution == 3 then
        // DU
        (u,gout) := Discrete.DiscreteUniform(
              g,
              p1,
              p2);
      elseif distribution == 4 then
        // Bin
        (u,gout) := Discrete.Binomial(
              g,
              p1,
              p2);
      elseif distribution == 5 then
        // Geom
        (u,gout) := Discrete.Geometric(
              g,
              p1);
      elseif distribution == 6 then
        // NegBin
        (u,gout) := Discrete.NegativeBinomial(
              g,
              p1,
              p2);
      elseif distribution == 7 then
        // Poisson
        (u,gout) := Discrete.Poisson(
              g,
              p1);
      elseif distribution == 8 then
        // U
        (u,gout) := Continuous.Uniform(
              g,
              p1,
              p2);
      elseif distribution == 9 then
        // Exp
        (u,gout) := Continuous.Exponential(
              g,
              p1);
      elseif distribution == 10 then
        // Erlang
        (u,gout) := Continuous.Erlang(
              g,
              p1,
              p2);
      elseif distribution == 11 then
        // Gamma
        (u,gout) := Continuous.Gamma(
              g,
              p1,
              p2);
      elseif distribution == 12 then
        // Weibull
        (u,gout) := Continuous.Weibull(
              g,
              p1,
              p2);
      elseif distribution == 13 then
        // N
        (u,gout) := Continuous.Normal(
              g,
              p1,
              p2);
      elseif distribution == 14 then
        // LN
        (u,gout) := Continuous.LogNormal(
              g,
              p1,
              p2);
      elseif distribution == 15 then
        // Beta
        (u,gout) := Continuous.Beta(
              g,
              p1,
              p2);
      elseif distribution == 16 then
        // Johnson
        (u,gout) := Continuous.Johnson(
              g,
              p1,
              p2,
              p3,p4);
      elseif distribution == 17 then
        // Triang
        (u,gout) := Continuous.Triangular(
              g,
              p1,
              p2,
              p3);
      else
        u := 0;
        gout := initGenerator();
      end if;
      Cg1 := gout.Cg[1];
      Cg2 := gout.Cg[2];
      Cg3 := gout.Cg[3];
      Cg4 := gout.Cg[4];
      Cg5 := gout.Cg[5];
      Cg6 := gout.Cg[6];
      Bg1 := gout.Bg[1];
      Bg2 := gout.Bg[2];
      Bg3 := gout.Bg[3];
      Bg4 := gout.Bg[4];
      Bg5 := gout.Bg[5];
      Bg6 := gout.Bg[6];
      Ig1 := gout.Ig[1];
      Ig2 := gout.Ig[2];
      Ig3 := gout.Ig[3];
      Ig4 := gout.Ig[4];
      Ig5 := gout.Ig[5];
      Ig6 := gout.Ig[6];
      Anti := gout.Anti;
      IncPrec := gout.IncPrec;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
GenerateVariate
</font></h3>
<p>
 
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>GenerateVariate</b>(Cg1,Cg2,Cg3,Cg4,Cg5,Cg6,Bg1,Bg2,Bg3,Bg4,Bg5,Bg6,Ig1,Ig2,Ig3,Ig4,Ig5,Ig6,Anti,IncPrec,distribution,p1,p2,p3,p4)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
This function generates a random variate from the distribution indicated in with the integer pararameter \"distribution\", using the following value correspondence:
<ul>
<li> 1 = Constant
<li> 2 = Bernoulli
<li> 3 = Discrete Uniform
<li> 4 = Binomial
<li> 5 = Geometric
<li> 6 = Negative Binomial
<li> 7 = Poisson
<li> 8 = Uniform
<li> 9 = Exponential
<li> 10 = Erlang
<li> 11 = Gamma
<li> 12 = Weibull
<li> 13 = Normal
<li> 14 = LogNormal
<li> 15 = Beta
<li> 16 = Johnson (bounded and unbounded)
<li> 17 = Triangular
<li> other value = initalization of the generator (call to initGenerator() function).
</ul>
 
Using this function, the generation of random variates in models can be easily parametrized. 
</P>
<p>
This function receives as parameters the individual components of a CMRG generator, instead of the record like in the GenerateVariate function.<br>
It returns the updated state of the generator also with its individual components.
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 (u,g.Cg1,g.Cg2,g.Cg3,g.Cg4,g.Cg5,g.Cg6,g.Bg1,g.Bg2,g.Bg3,g.Bg4,g.Bg5,g.Bg6,g.Ig1,g.Ig2,g.Ig3,g.Ig4,g.Ig5,g.Ig6,g.Anti,g.IncPrec) := <br> GenerateVariate(g.Cg1,g.Cg2,g.Cg3,g.Cg4,g.Cg5,g.Cg6,g.Bg1,g.Bg2,g.Bg3,g.Bg4,g.Bg5,g.Bg6,g.Ig1,g.Ig2,g.Ig3,g.Ig4,g.Ig5,g.Ig6,g.Anti,g.IncPrec,2,0.4);
 // Generates a discrete variate (u) with Bernoulli(0.4) probability distribution.
 
 (u,g.Cg1,g.Cg2,g.Cg3,g.Cg4,g.Cg5,g.Cg6,g.Bg1,g.Bg2,g.Bg3,g.Bg4,g.Bg5,g.Bg6,g.Ig1,g.Ig2,g.Ig3,g.Ig4,g.Ig5,g.Ig6,g.Anti,g.IncPrec,) := <br> GenerateVariate(g.Cg1,g.Cg2,g.Cg3,g.Cg4,g.Cg5,g.Cg6,g.Bg1,g.Bg2,g.Bg3,g.Bg4,g.Bg5,g.Bg6,g.Ig1,g.Ig2,g.Ig3,g.Ig4,g.Ig5,g.Ig6,g.Anti,g.IncPrec,11,0.4,0.6);
 // Generates a discrete variate (u) with Gamma(0.4,0.6) probability distribution.
 
The example <a href=\"Modelica://DESLib.RandomLib.Examples.Simple.VariatesSimple2\">VariatesSimple2</a> uses of this function.
 
 
 
</pre>
 
</HTML>
"));
    end GenerateVariateWrap;

    function Constant "Constant value generation"
      extends Var;
    algorithm
      x := a;
      gout := g;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Constant
</font></h3>
<p>
Generates a constant number.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Constant</b>(g,a)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Always generates the same number, equal to the parameter a.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
Constant(g,1) = 
(1,g)
 
for i in 1:5 loop
  (u[i],g) := Constant(g,3);
end for;
// u = {3,3,3,3,3};
 
</pre>
 
</HTML>
"));
    end Constant;

    package Discrete "Discrete probability distribution functions"
      function Discrete "Discrete"
        input CMRG.RngStream g "Random number generator";
        input Real p[:] "Cummulative probabilities";
        input Real v[:] "Values related with the probabilities";
        output Real value "Generated random variate";
        output CMRG.RngStream gout "Updated random number generator";
      protected
        Real u;
        Integer s = size(p,1);
        Boolean out;
        Integer i;
      algorithm
        (u,gout) := U01(g);
        out := false;
        i := 1;
        while not out loop
          if (u < p[i]) then
            value := v[i];
            out := true;
          end if;
          i := i +1;
        end while;
      /*  for i in 1:s loop
    if (u < p[i]) then
      value := v[i];
      u := 10;
    end if;
  end for;*/
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Discrete
</font></h3>
<p>
Empirical discrete probability distribution.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Discrete</b>(g,p[:],v[:])</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
This function generates random variates following the Empirical Discrete distribution function described by the parameters p and v.<br>
Parameter p represents the list of cummulative probabilities related with the values in v.<br>
Length of p and v must be the same.
<br><br>
The following figure graphically shows the empirical discrete distribution.
</p>
<p>
<IMG SRC=\"../Figs/discrete.png\" ALT=\"discrete\">
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 Discrete(g,{0.2,0.4,0.6,0.8,1},{1,2,3,4,5});
</pre>
 
</HTML>
"));
      end Discrete;

      function Bernoulli "Bernoulli"
        extends Var;
      algorithm
        (u,gout) := U01(g);
        if (u <= a) then
          x := 1;
        else
          x := 0;
        end if;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Bernoulli
</font></h3>
<p>
Bernoulli probability distribution.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Bernoulli</b>(g,p)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Bernoulli probability distribution with probability p.
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate u as U(0,1).<br>
2. If u &le p, return x = 1. Otherwise, return x = 0.
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(u,g) := Bernoulli(g,0.4);
 
</pre>
 
</HTML>
"));
      end Bernoulli;

      function DiscreteUniform "Discrete Uniform"
        extends Var;
      algorithm
        (u,gout) := U01(g);
        x := a + floor((b - a + 1)*u);
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
DiscreteUniform
</font></h3>
<p>
Discrete Uniform probability distribution.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>DiscreteUniform</b>(g,i,j)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Discrete Uniform probability distribution in the interval [i,j].
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate u as U(0,1).<br>
2. Return x = i + floor((j - i + 1)*u)
</p>
 
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
(u,g) := DiscreteUniform(g,2,4);
</pre>
 
</HTML>
"));
      end DiscreteUniform;

      function Binomial "Binomial"
        extends Var;
      algorithm
        x := 0;
        gout := g;
        for i in 1:a loop
          (u,gout) := Bernoulli(
                  gout,
                  b,
                  0,
                  0);
          x := x + u;
        end for;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Binomial
</font></h3>
<p>
Binomial probability distribution. 
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Binomial</b>(g,t,b)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following a Binomial(t,p) probability distribution. This distribution is the sum of t random variables from the Bernoulli(b) distribution.
</P>
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate y[1],...,y[t] as IID Bernoulli(p).<br>
2. Return x = y[1]+...+y[t]
</p>
 
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(u,g) := Binomial(g,4,0.3); 
 
</pre>
 
</HTML>
"));
      end Binomial;

      function Geometric "Geometric"
        extends Var;
      protected
        Real e;
        Real i;
      algorithm
        i := 0;
        e := 0;
        gout := g;
        while (e == 0) loop
          (u,gout) := U01(gout);
          if (u <= a) then
            x := i;
            e := 1;
          else
            i := i + 1;
          end if;
        end while;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Geometric
</font></h3>
<p>
Geometric probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Geometric</b>(g,b)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Geometric probability distribution with probabitity b.
</P>
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Set i = 0.<br>
2. Generate u as U(0,1).<br>
3. If u &le b, return i. Otherwise i=i+1 and go back to 2.
</p>
 
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
(u,g) := Geometric(g,0.5);
</pre>
 
</HTML>
"));
      end Geometric;

      function NegativeBinomial "Negative Binomial"
        extends Var;
      algorithm
        x := 0;
        gout := g;
        for i in 1:a loop
          (u,gout) := Geometric(gout,b);
          x := x + u;
        end for;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
NegativeBinomial
</font></h3>
<p>
Negative Binomial probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>NegativeBinomial</b>(g,t,b)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Negative Binomial probability distribution function. This variate is equivalent to the sum of t random variables following the Geometric(b) distribution.
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate y[1],...,y[t] as IID Geometric(p).<br>
2. Return x = y[1]+...+y[t]
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
(u,g) := Negativebinomial(g,6,0.2);
 
</pre>
 
</HTML>
"));
      end NegativeBinomial;

      function Poisson "Poisson"
        extends Var;
      protected
        Real aux;
        Real bux;
      algorithm
        aux := Modelica.Math.exp(-a);
        bux := 1;
        x := 0;
        gout := g;
        (u,gout) := U01(gout);
        bux := bux*u;
        while (bux >= aux) loop
          (u,gout) := U01(gout);
          bux := bux*u;
          x := x + 1;
        end while;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Poisson
</font></h3>
<p>
Poisson probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Poisson</b>(g,lambda)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Poisson probability distribution function with mean alpha.
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Let a = e(-lambda), b = 1, and i = 0.<br>
2. Generate u as U(0,1), and replace b = b*u. If b &lt a, return x = i. Otherwise, go to step 3.<br>
3. Replace i by i+1 and go to step 2.
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(u,g) := Poisson(4.5); 
</pre>
 
</HTML>
"));
      end Poisson;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<p>
The <b>Discrete</b> package contains functions for generating random variates following several discrete probability distributions.
</p>
 
<p>
<dl>
<dt><b>Author:</b></dt>
<dd><a href=\"http://www.euclides.dia.uned.es/vsanz\"> Victorino Sanz </a><br>
    Dpto. Inform&aacute;tica y Autom&aacute;tica, UNED<br>
    Juan del Rosal, 16<br>
    28040, Madrid<br>
    Spain<br>
    email: <A HREF=\"mailto:vsanz@dia.uned.es\">vsanz@dia.uned.es</A><br></dd>
</dl>
</p>
 
<p>
<dt><b>Copying:</b></dt>
<p>
<p>
Licensed by Victorino Sanz under the Modelica License 2 <br>
Copyright 2009, Victorino Sanz.
</p>
<p>
<i>This Modelica packages is <u>free</u> software and the use is completely at <u>your ouwn risk;</u> it can be redistributed and/or modified under the terms of the Modelica license 2, 
see the license conditions (including the disclaimer of warranty) <a href=\"Modelica://DESLib.ModelicaLicense2\">here</a> or at <a href=\"http://www.modelica.org/licenses/ModelicaLicense2\">http://www.modelica.org/licenses/ModelicaLicense2</a>
</p><br>
</HTML>
"));
    end Discrete;

    package Continuous "Continuous probability distribution functions"
      function Continuous "Empirical Continuous"
        input CMRG.RngStream g;
        input Real p[:];
        input Real v[:];
        output Real value;
        output CMRG.RngStream gout;
      protected
        Real u;
        Integer s = size(p,1);
      algorithm
        (u,gout) := U01(g);
        for i in 1:s loop
          if (u <= p[i]) then
            if i == 1 then
              value := v[1];
            else
              value := v[i-1] + (((u-p[i-1])*(v[i]-v[i-1]))/(p[i]-p[i-1]));
            end if;
            u := 10;
          end if;
        end for;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Continuous
</font></h3>
<p>
Empirical Continuous probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Continuous</b>(g,p[:],v[:])</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
This function generates random variates following the Empirical Continuous distribution function described by the parameters p and v.<br>
Parameter p represents the list of cummulative probabilities related with the values in v.<br>
Length of p and v must be the same.
</P>
The following figure graphically shows the empirical discrete distribution.
</p>
<p>
<IMG SRC=\"../Figs/continuous.png\" ALT=\"continuous\">
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 Continuous(g,{0.2,0.4,0.6,0.8,1},{1,2,3,4,5});
 
 
</pre>
 
</HTML>
"));
      end Continuous;

      function Uniform "Uniform"
        extends Var;
      protected
        Generator gaux;
      algorithm
        (u,gout) := U01(g);
        x := a + (b - a)*u;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Uniform
</font></h3>
<p>
Uniform probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Uniform</b>(g,a,b)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Uniform probability distribution function in the interval [a,b].
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate u as U(0,1)<br>
2. Return x = a + (b-a)*u
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(u,g) := Uniform(g,2,4);
</pre>
 
</HTML>
"));
      end Uniform;

      function Exponential "Exponential"
        extends Var;
      algorithm
        (u,gout) := U01(g);
        x := (-a)*Modelica.Math.log(1-u);
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Exponential
</font></h3>
<p>
Exponential probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Exponential</b>(g,beta)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Exponential probability distribution of mean beta.
</P>
 
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate u as U(0,1)<br>
2. Return x = -beta * ln(1-u)
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
(u,g) := Exponential(g,0.4);
</pre>
 
</HTML>
"));
      end Exponential;

      function Erlang "Erlang"
        extends Var;
      protected
        Integer k =  integer(b);
        Real y[k];
      algorithm
        gout := g;
        for i in 1:k loop
          (y[i],gout) := U01(gout);
        end for;
        x := (-a/b)*Modelica.Math.log(product(y));
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Erlang
</font></h3>
<p>
Erlang probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Erlang</b>(g,beta,m)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Erlang probability distribution function, with parameters beta and m. The generated variate corresponds to the sum of m random variates following the Exponential distribution of mean beta/m.
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate y[1],...,y[m] as IID U(0,1)<br>
2. Return x = (-beta/m)*ln(product(y))
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
  (u,g) := Erlang(g,0.5,4); 
</pre>
 
</HTML>
"));
      end Erlang;

      function Gamma "Gamma"
        extends Var;
      protected
        Real Y;
        Real y;
        Real y1;
        Real b1;
        Real p;
        Real u1;
        Real u2;
        Boolean step1;
        Boolean step2;
        Boolean step3;
        Boolean step4;
        Real ca;
        Real cb;
        Real q;
        Real theta;
        Real cd;
        Real V;
        Real Z;
        Real W;
      algorithm
        gout := g;
        if a == 1 then //<= 1 and a >= 1 then
          (y,gout) := Exponential(gout,a);
        elseif a == 0.5 then //<= 0.5 and a>= 0.5 then
          (Y,gout) := Normal(gout,0,1);
          y := Y^2;
        elseif a < 1 then
          b1 := (Modelica.Constants.e + a)/Modelica.Constants.e;
          step1 := true;
          while step1 loop
            (u1,gout) := U01(gout);
            p := b1*u1;
            if p > 1 then
              step3 := true;
              step2 := false;
            else
              step3 := false;
              step2 := true;
            end if;
            if step2 then
              Y := p^(1/a);
              (u2,gout) := U01(gout);
              if u2 <= Modelica.Constants.e^(-y1) then
                step1 := false;
                y := Y;
              end if;
            end if;
            if step3 then
              Y := -Modelica.Math.log((b1-p)/a);
              (u2,gout) := U01(gout);
              if u2 <= Y^(a-1) then
                step1 := false;
                y := Y;
              end if;
            end if;
          end while;
        else
          ca := 1/(sqrt(2*a -1));
          cb := a - Modelica.Math.log(4);
          q := a+1;
          theta := 4.5;
          cd := 1 + Modelica.Math.log(theta);
          step1 := true;
          while step1 loop
            (u1,gout) := U01(gout);
            (u2,gout) := U01(gout);
            V := ca*Modelica.Math.log(u1/(1-u1));
            Y := a*(Modelica.Constants.e^V);
            Z := (u1^2)*u2;
            W := cb + q*V - Y;
            step4 := true;
            if (W+cd-(theta*Z) >= 0) then
              step1 := false;
              step4 := false;
              y := Y;
            end if;
            if step4 then
              if W >= Modelica.Math.log(Z) then
                step1 := false;
                y := Y;
              end if;
            end if;
          end while;
        end if;
        x := b*y; // Gamma(a,b) = b * Gamma(a,1)
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Gamma
</font></h3>
<p>
Gamma probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Gamma</b>(g,alpha,beta)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Gamma probability distribution function, with parameters alpha and beta.
</P>
 
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
 
I. If alpha == 1, return y = exponential(1)<br>
II. If alpha == 0.5, return y = normal(0,1)<br>
III. If a &lt 1<br>
&nbsp &nbsp &nbsp &nbsp  b = (e + alpha)/e<br>
&nbsp &nbsp &nbsp &nbsp 1. Generate u1 as U(0,1). Let P = beta*u1. If P > 1, go to step 3. Otherwise, go to step 2.<br>
&nbsp &nbsp &nbsp &nbsp 2. Let Y = P^(1/alpha), and generate u2 as U(0,1). If u2 &le e^(-Y), return y = Y. Otherwise, go to step 1.<br>
&nbsp &nbsp &nbsp &nbsp 3. Let Y = -ln[(b - P)/alpha] and generate u2 as U(0,1). If u2 &le Y^(alpha-1), return y = Y. Otherwise, go to step 1.<br>
IV. If a &gt 1
&nbsp &nbsp &nbsp &nbsp a = 1/sqrt(2*alpha -1)<br>
&nbsp &nbsp &nbsp &nbsp b = alpha -ln(4)<br>
&nbsp &nbsp &nbsp &nbsp q = alpha + 1/a<br>
&nbsp &nbsp &nbsp &nbsp theta = 4.5<br>
&nbsp &nbsp &nbsp &nbsp d = 1 + ln(theta)<br>
&nbsp &nbsp &nbsp &nbsp 1. Generate u1 and u2 as IID U(0,1).<br>
&nbsp &nbsp &nbsp &nbsp 2. Let V = a*ln[u1/(1-u1)], Y = alpha*e^V, Z = u1^2*u2, and W = b + q*V - Y.<br>
&nbsp &nbsp &nbsp &nbsp 3. If W + d - theta*Z &ge 0, return y = Y. Otherwise, go to step 4.<br>
&nbsp &nbsp &nbsp &nbsp 3. If W &ge ln(Z), return y = Y. Otherwise, go to step 1.<br>
V. x = beta*y;
 
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
  (u,g) := Gamma(g,0.4,0.7);
</pre>
 
</HTML>
"));
      end Gamma;

      function Weibull "Weibull"
        extends Var;
      protected
        Real u;
      algorithm
        (u,gout) := U01(g);
        x := a*((-Modelica.Math.log(1-u))^(1/b));
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Weibull
</font></h3>
<p>
Weibull probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Weibull</b>(g,alpha,beta)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Weibull probability distribution, with parameters alpha and beta.
</P>
 
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate u as U(0,1)<br>
2. Return x = alpha*((-ln(1-u))^(1/beta))
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
(u,g) := Weibull(g,0.4,0.8);
 
</pre>
 
</HTML>
"));
      end Weibull;

      function Normal "Normal"
        extends Var;
      protected
        Real p;
        constant Real a0 = 2.50662823884;
        constant Real a1 = -18.61500062529;
        constant Real a2 = 41.39119773534;
        constant Real a3 = -25.44106049637;
        constant Real b1 = -8.47351093090;
        constant Real b2 = 23.08336743743;
        constant Real b3 = -21.06224101826;
        constant Real b4 = 3.13082909833;
        constant Real c0 = -2.78718931138;
        constant Real c1 = -2.29796479134;
        constant Real c2 = 4.85014127135;
        constant Real c3 = 2.32121276858;
        constant Real d1 = 3.54388924762;
        constant Real d2 = 1.63706781897;
        Real r;
        constant Real split = 0.42;
        Real value;
      algorithm
        gout:= g;
        (p,gout) := U01(gout);
        if (abs(p-0.5) <= split) then
          r := (p-0.5) *(p-0.5);
          value := (p-0.5)*(((a3*r+a2)*r+a1)*r+a0)/((((b4*r+b3)*r+b2)*r+b1)*r+1);
        elseif (0 < p and p < 1) then
          if (0.5 < p) then
            r := sqrt(-Modelica.Math.log(1-p));
          else
            r := sqrt(-Modelica.Math.log(p));
          end if;
          value :=(((c3*r + c2)*r + c1)*r + c0)/((d2*r + d1)*r + 1);
          if (p < 0.5) then
            value :=-value;
          end if;
        else
          value := 0;
        end if;
        x := (value*b)+a;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Normal
</font></h3>
<p>
Normal probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Normal</b>(g,mu,sigma2)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Normal probability distribution, with mean mu and variance sigma2.
</P>
 
<p>
The algorithm used for this distribution is the same that the one implemented in SIMAN, and is detailed in:<br><br>
J.D. Beasley and S.G. Springer (1977), The Percentage Points of the Normal Distribution, <i>Applied Statistics</i>, vol. 26, pp: 118-121.
 
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 
(u,g) := Normal(g,0,1);
 
</pre>
 
</HTML>
"));
      end Normal;

      function LogNormal "LogNormal"
        extends Var;
      protected
        Real y;
      algorithm
        //gout := g;
        (y,gout) := Normal(g,a,b);
        x := Modelica.Math.exp(y);
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
LogNormal
</font></h3>
<p>
LogNormal probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>LogNormal</b>(g,mu,sigma2)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the LogNormal probability distribution function, with mean mu and variance sigma2.
</P>
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate Y as N(mu,sigma2).<br>
2. Return x = e^Y
</p>
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(u,g) := LogNormal(g,0,1);
</pre>
 
</HTML>
"));
      end LogNormal;

      function Beta "Beta"
        extends Var;
      protected
        Real y1;
        Real y2;
      algorithm
        (y1,gout) := Gamma(g,a,1);
        (y2,gout) := Gamma(gout,b,1);
        x := y1/(y1+y2);
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Beta
</font></h3>
<p>
Beta probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Beta</b>(g,alpha1,alpha2)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Beta probability distribution, with parameters alpha1 and alpha2.
</P>
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate Y1 as gamma(alpha1,1), and Y2 as gamma(alpha2,1) independent from Y1.<br>
2. Return x = Y1/(Y1+Y2)
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
(u,g) := Beta(g,0.6,0.4); 
</pre>
 
</HTML>
"));
      end Beta;

      function Johnson "Johnson"
      extends Var;
      protected
        Real z;
        Real y;
      algorithm
        if (b > 0) then // bounded
          (z,gout) := Normal(g,0,1);
          y := Modelica.Math.exp((z -a)/b);
          x := (c + d*y)/(y+1);
        else // unbounded
          (z,gout) := Normal(g,0,1);
          y := Modelica.Math.exp((z-a)/b);
          x := c + (d/2)*(y-1/y);
        end if;
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Johnson
</font></h3>
<p>
Johnson probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Johnson</b>(g,alpha1,alpha2,a,b) if bounded (alpha2 > 0) </pre></blockquote> 
<blockquote><pre><b>Johnson</b>(g,alpha1,alpha2,gamma,beta) if unbounded (alpha2 &lt; 0)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Johnson probability distribution, with parameters alpha1 and alpha2. If alpha2 > 0, for the Johnson bounded distribution, the variate is generated in the interval [a,b]. Otherwise, gamma and beta are the shape parameters.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
 (u,g) := Johnson(g,0.1,0.5,1,4); // bounded
 (u,g) := Johnson(g,0.1,-0.5,1,1); // unbounded
</pre>
 
</HTML>
"));
      end Johnson;

      function Triangular "Triangular"
        extends Var;
      protected
        Real cc;
      algorithm
        (u,gout) := U01(g);
        cc := (b-a)/(c-a);
        if u < cc then
          x := sqrt(cc*u);
        else
          x := 1-sqrt((1-cc)*(1-u));
        end if;
        x := a + (c-a)*x;
      /*  if (u < ((b - a)/(c - a))) then
    x := a + sqrt(u*(c - a)*(b - a));
    //x := a + sqrt(u*((b - a)/(c - a)));
  else
    x := c - sqrt((1 - u)*(c - a)*(c - b));
  end if;*/
      annotation(preferredView="info",
        Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
Triangular
</font></h3>
<p>
Triangular probability distribution function.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>Triangular</b>(g,min,mode,max)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Generates a random variate following the Triangular probability distribution function. The variate is generated in the interval [min,max] with the desired mode.
</P>
<p>
The pseudo-code algorithm used for this distribution is:<br><br>
 
1. Generate u as U(0,1)<br>
2. If u &le ((mode - min)/(max - min)), return x = min + sqrt(u*(max - min)*(mode - min)). Otherwise, x = max - sqrt((1-u)*(max - min)*(max - mode))
</p>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
  (u,g) := Triangular(g,1,3,6); 
</pre>
 
</HTML>
"));
      end Triangular;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<p>
The <b>Continuous</b> package contains functions to generate random variates following some continuous probability distributions.
</p>
 
<p>
<dl>
<dt><b>Author:</b></dt>
<dd><a href=\"http://www.euclides.dia.uned.es/vsanz\"> Victorino Sanz </a><br>
    Dpto. Informática y Automática, UNED<br>
    Juan del Rosal, 16<br>
    28040, Madrid<br>
    Spain<br>
    email: <A HREF=\"mailto:vsanz@dia.uned.es\">vsanz@dia.uned.es</A><br></dd>
</dl>
</p>
 
<p>
<dt><b>Copying:</b></dt>
<p>
<p>
Licensed by Victorino Sanz under the Modelica License 2 <br>
Copyright 2009, Victorino Sanz.
</p>
<p>
<i>This Modelica packages is <u>free</u> software and the use is completely at <u>your ouwn risk;</u> it can be redistributed and/or modified under the terms of the Modelica license 2, 
see the license conditions (including the disclaimer of warranty) <a href=\"Modelica://DESLib.ModelicaLicense2\">here</a> or at <a href=\"http://www.modelica.org/licenses/ModelicaLicense2\">http://www.modelica.org/licenses/ModelicaLicense2</a>
</p><br>
</HTML>
"));
    end Continuous;
    /*******************/
    /*********************/
    annotation(preferredView="info",
    Documentation(info="<HTML>
<p>
<b>Variates</b> Modelica package contains the implementation of random variate generation functions for the following probability distributions.<br>
<br>
Discrete Probability Distributions:
<ul>
<li> Empirical Discrete
<li> Bernoulli
<li> Discrete Uniform
<li> Binomial
<li> Geometric
<li> Negative Binomial
<li> Poisson
</ul>
 
Continuous Probability Distributions:
<ul>
<li> Empirical Continuous
<li> Uniform
<li> Exponential
<li> Erlang
<li> Gamma
<li> Weibull
<li> Normal
<li> LogNormal
<li> Beta
<li> Johnson (bounded and unbounded)
<li> Triangular
</ul>
</p>
 
<p> A description of the package structure and its use is given in the <a href=\"Modelica://DESLib.RandomLib.Variates.UsersGuide\">Users Guide</a>
 
<p>
<dl>
<dt><b>Author:</b></dt>
<dd><a href=\"http://www.euclides.dia.uned.es/vsanz\"> Victorino Sanz </a><br>
    Dpto. Inform&aacute;tica y Autom&aacute;tica, UNED<br>
    Juan del Rosal, 16<br>
    28040, Madrid<br>
    Spain<br>
    email: <A HREF=\"mailto:vsanz@dia.uned.es\">vsanz@dia.uned.es</A><br></dd>
</dl>
</p>
 
<p>
<dt><b>Copying:</b></dt>
<p>
<p>
Licensed by Victorino Sanz under the Modelica License 2 <br>
Copyright 2009, Victorino Sanz.
</p>
<p>
<i>This Modelica package is <u>free</u> software and the use is completely at <u>your ouwn risk;</u> it can be redistributed and/or modified under the terms of the Modelica license 2, 
see the license conditions (including the disclaimer of warranty) <a href=\"Modelica://DESLib.ModelicaLicense2\">here</a> or at <a href=\"http://www.modelica.org/licenses/ModelicaLicense2\">http://www.modelica.org/licenses/ModelicaLicense2</a>
</p><br>
</HTML>
"));
  end Variates;

  package Sources
    model Discrete
      parameter Integer distribution = 1 "Random Distribution"
    annotation(choices(choice = 1 "Constant",
          choice=2 "Bernoulli",
          choice=3 "Discrete Uniform",
          choice=4 "Binomial",
          choice=5 "Geometric",
          choice=6 "Negative Binomial",
          choice=7 "Poisson"));
      parameter Real p1 = 1 "first parameter";
      parameter Real p2 = 1 "second parameter";
        Variates.Generator g "RNG";
      Modelica.Blocks.Interfaces.IntegerOutput y annotation (Placement(
            transformation(extent={{80,-10},{100,10}}, rotation=0)));
    protected
      Real out;
    algorithm
      // initilization of the RngStream
      when initial() then
        g := Variates.initGenerator();
      end when;
      (out,g) := Variates.GenerateVariate(distribution,g,p1,p2);
      y :=integer(out);
      annotation (Icon(graphics),
                        Diagram(graphics));
    end Discrete;

    model Continuous
      Variates.Generator g "RNG";
      parameter Integer distribution = 1 "Random distribution"
    annotation(choices(choice = 1 "Constant",
          choice=8 "Uniform",
          choice=9 "Exponential",
          choice=10 "Erlang",
          choice=11 "Gamma",
          choice=12 "Weibull",
          choice=13 "Normal",
          choice=14 "LogNormal",
          choice=15 "Beta",
          choice=16 "Johnson (bounded and unbounded)",
          choice=17 "Triangular"));
      parameter Real p1 = 1;
      parameter Real p2 = 0.1;
      parameter Real p3 = 1;
      parameter Real p4 = 1;
      Modelica.Blocks.Interfaces.RealOutput y annotation (Placement(
            transformation(extent={{80,-10},{100,10}}, rotation=0)));
    algorithm
      // initilization of the RngStream
      when initial() then
        g := Variates.initGenerator();
      end when;
      (y,g) := Variates.GenerateVariate(distribution,g,p1,p2,p3,p4);
      annotation (Diagram(graphics));
    end Continuous;
  end Sources;

  package CMRG
    "Combined Multiple Recursive Generator for uniform random numbers generation"
    model UsersGuide "Users Guide"
        extends Modelica.Icons.Information;

        annotation (Documentation(info="<html>
<h3><font color=\"#008000\" size=5>Users Guide of the CMRG Package</font></h3>
 
<p>
<b>CMRG</b> package is the implementation in Modelica of the RNG algorithm proposed by L'Ecuyer et al.
(<a href=\"http://www.iro.umontreal.ca/~lecuyer/myftp/papers/streams00.pdf\">pdf</a>)<br>
This RNG algorithm belongs to the Combined Multiple Recursive Generators family and combines two recursive generators of order 3.<br>
</p>
 
<p>
The RNG implemented in CMRG package has a period close to 2^191.<br> 
This generator gives the possibility of creating multiple streams and sub-streams than can be considered as independent RNGs.<br>
The whole period of the generator can be divided into disjoint streams, each of length 2^127.
At the same time, each stream can be divided into 2^51 adjacent sub-streams, each of length 2^76.
</p>
 
<p>
The main class in the CMRG package is the <em>RngStream</em>. 
It represents a single stream of sequential random numbers.<br> 
As mentioned above, an RngStream is divided into adjacent sub-streams.
The state of the RngStream indicates the start, current and next position in the current stream and sub-stream.
</p> 
 
<h3><font color=\"#008000\">Package Structure</font></h3>
<p>
 The CMRG package is composed of: 1) the developer package, called Src; and 2) the set of public functions.<br>
The developer package contains the private part of the package. This package includes contants and functions to perform the required calculations to manage the state of the RngStreams. They have no utility for an standard user.<br>
The public functions describe the operations that a user can perform with an RngStream, such as generating random numbers, updating the state of the stream or changing the package seed among others.<br>
The operation of each function is described in its documentation.
</p>
 
<h3><font color=\"#008000\">Random Number Generation</font></h3>
<p>
In order to generate uniform random numbers using CMRG package the user has to:
<ul>
<li> Declare an RngStream that represents the sequential stream of random numbers.
<li> Initialize the RngStream using the CreateStream function.
<li> Call the RandU01 function to generate a random number. This function will return the generated random number and the updated state of the RngStream.
</ul>
 
Notice that the updated state of the RngStream has to be used to generate the next random number.<br>
</p>
<p>
The following example represents the steps indicated above.
 
<pre>
model usingCMRG 
  // declaration of the streams 
  CMRG.RngStream g1; 
  Real u1[6];
algorithm 
  when initial() then
    // stream initialization  
    g1 := CMRG.RngStream_CreateStream();
  end when;
  when time &lt;= 0 then
    for i in 1:6 loop
      // generation of random numbers from g1
      (u1[i],g1):= CMRG.RngStream_RandU01(g1);
    end for;
  end when;
end usingCMRG;
</pre>
</p>
 
<p>
Several independent sources of random numbers can be created by declaring several RngStreams. The initialization of each declared RngStream will give it a different initial state.<br>
The package seed is automatically managed, so it is not required to set an initial seed.
 
</p> 
</html>"));
    end UsersGuide;

    package Src "Internal data and functions"
    type Seed = Real[6] "RNG seed data type";
    constant Real norm =  2.328306549295727688e-10;
    constant Real m1 =    4294967087.0;
    constant Real m2 =    4294944443.0;
    constant Real a12 =     1403580.0;
    constant Real a13n =     810728.0;
    constant Real a21 =      527612.0;
    constant Real a23n =    1370589.0;
    constant Real two17 =   131072.0;
    constant Real two53 =   9007199254740992.0;
    constant Real fact =  5.9604644775390625e-8; /* 1 / 2^24 */
    /* Default initial seed of the package. Will be updated to become
     the seed of the next created stream. */
      //constant Real[6] nextSeed = {12345, 12345, 12345, 12345, 12345, 12345};
    /* The following are the transition matrices of the two MRG components */
    /* (in matrix form), raised to the powers -1, 1, 2^76, and 2^127, resp.*/
    /* Inverse of A1p0 */
    constant Real[3,3] InvA1 = {
            { 184888585.0,   0.0,  1945170933.0},
            {         1.0,   0.0,           0.0},
            {         0.0,   1.0,           0.0}};
    /* Inverse of A2p0 */
    constant Real[3,3] InvA2 =  {
            {      0.0,  360363334.0,  4225571728.0},
            {      1.0,          0.0,           0.0},
            {      0.0,          1.0,           0.0}};
    constant Real[3,3] A1p0 = {
            {       0.0,        1.0,       0.0},
            {       0.0,        0.0,       1.0},
            { -810728.0,  1403580.0,       0.0}};
    constant Real[3,3] A2p0 =  {
            {        0.0,        1.0,       0.0},
            {        0.0,        0.0,       1.0},
            { -1370589.0,        0.0,  527612.0}};
    constant Real[3,3] A1p76= {
            {      82758667.0, 1871391091.0, 4127413238.0},
            {    3672831523.0,   69195019.0, 1871391091.0},
            {    3672091415.0, 3528743235.0,   69195019.0}};
    constant Real[3,3] A2p76= {
            {    1511326704.0, 3759209742.0, 1610795712.0},
            {    4292754251.0, 1511326704.0, 3889917532.0},
            {    3859662829.0, 4292754251.0, 3708466080.0}};
    constant Real[3,3] A1p127 = {
            {    2427906178.0, 3580155704.0,  949770784.0},
            {     226153695.0, 1230515664.0, 3580155704.0},
            {    1988835001.0,  986791581.0, 1230515664.0}};
    constant Real[3,3] A2p127 = {
            {    1464411153.0,  277697599.0, 1610723613.0},
            {      32183930.0, 1464411153.0, 1022607788.0},
            {    2824425944.0,   32183930.0, 2093834863.0}};
    /************** from headers file *******************/
    /****************************************************/
    /*-------------------------------------------------------------------------*/
    function MultModM
        "Compute (a*s + c) % m. m must be < 2^35.  Works also for s, c < 0"
      input Real a;
      input Real s;
      input Real c;
      input Real m;
      output Real v;
      protected
      Real a1;
      Real aux;
    algorithm
      aux := a;
      v := aux * s + c;
      if ((v >= two53) or (v <= -two53)) then
        a1 := floor( aux / two17);
        aux := aux - (a1 * two17);
        v := a1 * s;
        a1 := floor(v / m);
        v := v - (a1 * m);
        v := v * two17 + aux * s + c;
      end if;
      a1 := floor( v / m);
      v := v - (a1*m);
      if (v < 0.0) then
        v := v + m;
      end if;
    end MultModM;
    /*-------------------------------------------------------------------------*/
    function MatVecModM
        "MatVecModM returns v = A*s % m.  Assumes that -m < s[i] < m. Works even if v = s."
      input Real[3,3] A;
      input Real[3] s;
      output Real[3] v;
      input Real m;
      protected
      Real[3] x;
    algorithm
      for i in 1:3 loop
        x[i] := MultModM( A[i,1], s[1], 0.0, m);
        x[i] := MultModM( A[i,2], s[2], x[i], m);
        x[i] := MultModM( A[i,3], s[3], x[i], m);
      end for;
      for i in 1:3 loop
        v[i] := x[i];
      end for;
    end MatVecModM;
    /*-------------------------------------------------------------------------*/
    function MatMatModM
        "MatMAtModM Returns C = A*B % m. Work even if A = C or B = C or A = B = C."
      input Real[3,3] A;
      input Real[3,3] B;
      input Real m;
      output Real[3,3] C;
      protected
      Real[3] V;
      Real[3,3] W;
    algorithm
      for i in 1:3 loop
        for j in 1:3 loop
          V[j] := B[j,i];
        end for;
        V := MatVecModM(A, V, m);
        for j in 1:3 loop
          W[j,i] := V[j];
        end for;
      end for;
      for i in 1:3 loop
        for j in 1:3 loop
          C[i,j] := W[i,j];
        end for;
      end for;
    end MatMatModM;
    /*-------------------------------------------------------------------------*/
    function MatTwoPowModM
        "MatTwoPowModM compute matrix B = (A^(2^e) % m);  works even if A = B"
      input Real[3,3] A;
      output Real[3,3] B;
      input Real m;
      input Real e;
      protected
      Integer i;
      Integer j;
    algorithm
      /* initialize: B = A */
      //if (A[1,1] <> B[1,1]) or (A[1,2] <> B[1,2]) or (A[1,3] <> B[1,3]) or (A[2,1] <> B[2,1]) or (A[2,2] <> B[2,2]) or (A[2,3] <> B[2,3]) or (A[3,1] <> B[3,1]) or (A[3,2] <> B[3,2]) or (A[3,3] <> B[3,3]) then
        for i in 1:3 loop
          for j in 1:3 loop
             B[i,j] := A[i,j];
          end for;
        end for;
      //end if;
      /* Compute B = A^{2^e} */
      for i in 1:e loop
        B := MatMatModM( B, B, m);
      end for;
    end MatTwoPowModM;
    /*-------------------------------------------------------------------------*/
    function MatPowModM
        "MatPowModM compute matrix B = A^n % m ;  works even if A = B"
      input Real[3,3] A;
      output Real[3,3] B;
      input Real m;
      input Real n;
      protected
      Real[3,3] W;
      Real naux;
      Real mo;
    algorithm
      /* initialize: W = A; B = I */
      for i in 1:3 loop
        for j in 1:3 loop
          W[i,j] := A[i,j];
          B[i,j] := 0.0;
        end for;
      end for;
      for j in 1:3 loop
        B[j,j] := 1.0;
      end for;
      naux := n;
      /* Compute B = A^n % m using the binary decomposition of n */
      while (naux > 0) loop
        mo := floor(mod(naux,2));
        if (mo <> 0) then
          B := MatMatModM( W, B, m);
        end if;
        W := MatMatModM( W, W, m);
        naux := naux/2;
      end while;
    end MatPowModM;
    /*-------------------------------------------------------------------------*/
    function U01 "Uniform random number generation function"
      input RngStream g "RngStream";
      output Real out "Uniform random number";
      output RngStream gout "Updated RngStream";
      protected
      Real k;
      Real p1;
      Real p2;
      Real u;
    algorithm
      gout := g;
      /* Component 1 */
      p1 := (a12 * gout.Cg[2]) - (a13n * gout.Cg[1]);
      k := floor(p1 / m1);
      p1 := p1 - ( k * m1);
      if (p1 < 0.0) then
        p1 := p1 + m1;
      end if;
      gout.Cg[1] := gout.Cg[2];
      gout.Cg[2] := gout.Cg[3];
      gout.Cg[3] := p1;
      /* Component 2 */
      p2 := (a21 * gout.Cg[6]) - (a23n * gout.Cg[4]);
      k := floor(p2 / m2);
      p2 := p2 - (k * m2);
      if (p2 < 0.0) then
        p2 := p2 + m2;
      end if;
      gout.Cg[4] := gout.Cg[5];
      gout.Cg[5] := gout.Cg[6];
      gout.Cg[6] := p2;
      /* Combination */
      if (p1 > p2) then
        u := (p1 -p2) * norm;
      else
        u := (p1 - p2 + m1) * norm;
      end if;
      if (gout.Anti <> 0) then
        out := 1-u;
      else
        out := u;
      end if;
    end U01;
    /*-------------------------------------------------------------------------*/
    function U01d
        "Uniform random number generation function, with increased precission"
      input RngStream g "RngStream";
      output Real out "Uniform random number";
      output RngStream gout "Updated RngStream";
      protected
      Real u;
      Real uaux;
    algorithm
      (u,gout) := U01(g);
      if (g.Anti == 0) then
        (uaux,gout) := U01(gout);
        u := u +( uaux * fact);
        if (u < 1.0) then
          out := u;
        else
          out := u - 1.0;
        end if;
      else
        /* Don't forget that U01() returns 1 - u in the antithetic case */
        (uaux,gout) := U01(gout);
        u :=u + ((uaux - 1.0)*fact);
        if (u < 0.0) then
          out := u + 1.0;
        else
          out := u;
        end if;
      end if;
    end U01d;
    /*-------------------------------------------------------------------------*/
    function CheckSeed "Checks correctness of a seed."
      input Seed seed "Seed to be checked";
      output Integer out "Results of the check";
      protected
      Integer i;
    algorithm
      /* Check that the seeds are legitimate values. Returns 0 if legal seeds,
       -1 otherwise */
      out := 0;
      for i in 1:3 loop
        if (seed[i] >= m1) then
          out := -1;
        end if;
      end for;
      for i in 4:6 loop
        if (seed[i] >= m2) then
          out
       := -1;
        end if;
      end for;
      if (seed[1] == 0) and (seed[2] == 0) and (seed[3] == 0) then
        out := -1;
      end if;
      if (seed[4] == 0) and (seed[5] == 0) and (seed[6] == 0) then
        out := -1;
      end if;
    end CheckSeed;

      function SaveSeed "saves seed to file CMRGSeed.txt"
        input Seed seed;
        output Integer out;
      //protected
      //  Integer lon;
      //  Seed aux;
      algorithm
      /*  (aux,lon) := ReadVar("seed", 6);
  if (lon == 0) then
    CreateVar("seed",seed,6);
  else
    UpdateVar("seed",seed,6);
  end if;
  out :=0;*/
        if Modelica.Utilities.Files.exist("CMRGSeed.txt") then
          Modelica.Utilities.Files.remove("CMRGSeed.txt");
          Modelica.Utilities.Streams.print(String(seed[1],significantDigits=14)+"\t"+String(seed[2],significantDigits=14)+"\t"+String(seed[3],significantDigits=14)+"\t"+String(seed[4],significantDigits=14)+"\t"+String(seed[5],significantDigits=14)+"\t"+String(seed[6],significantDigits=14),"CMRGSeed.txt");
        else
          Modelica.Utilities.Streams.print(String(seed[1],significantDigits=14)+"\t"+String(seed[2],significantDigits=14)+"\t"+String(seed[3],significantDigits=14)+"\t"+String(seed[4],significantDigits=14)+"\t"+String(seed[5],significantDigits=14)+"\t"+String(seed[6],significantDigits=14),"CMRGSeed.txt");
        end if;
        out := 1;
        Modelica.Utilities.Streams.close("CMRGSeed.txt");
      end SaveSeed;

      function ReadSeed
        "reads the current seed from file CMRGSeed.txt. If the file does not exists, saves the default value {12345,12345,12345,12345,12345,12345}"
        output Seed out;
      protected
        Integer aux=0;
        Integer found;
        String line = "";
        Integer index = 1;
      algorithm
        if Modelica.Utilities.Files.exist("CMRGSeed.txt") then
          line := Modelica.Utilities.Streams.readLine("CMRGSeed.txt",1);
          for i in 1:6 loop
            (out[i],index) := Modelica.Utilities.Strings.scanInteger(line,index,false,"");
          end for;
        else
          for i in 1:6 loop
            out[i] := 12345;
          end for;
          Modelica.Utilities.Streams.print(String(12345)+"\t"+String(12345)+"\t"+String(12345)+"\t"+String(12345)+"\t"+String(12345)+"\t"+String(12345),"CMRGSeed.txt");
        end if;
        Modelica.Utilities.Streams.close("CMRGSeed.txt");
      end ReadSeed;
    end Src;
    /*---------------------------------------------------------------------*/
    /* Public part.                                                        */
    /*---------------------------------------------------------------------*/
  record RngStream "Data structure for an RNG stream"
    Real[6] Cg;
    Real[6] Bg;
    Real[6] Ig;
    Integer Anti;
    Integer IncPrec;
  end RngStream;

    function CreateStream "CreateStream"
      output RngStream g "Created RngStream";
    protected
      Src.Seed seed;
      Src.Seed outseed;
      Real[3] A;
    algorithm
      seed :=  Src.ReadSeed();
      g.Anti := 0;
      g.IncPrec := 0;
      for i in 1:6 loop
        g.Bg[i] := seed[i];
        g.Cg[i] := seed[i];
        g.Ig[i] := seed[i];
      end for;
      for i in 1:3 loop
        A[i] := seed[i];
      end for;
      A := Src.MatVecModM( Src.A1p127, A, Src.m1);
      for i in 1:3 loop
        outseed[i] := A[i];
      end for;
      for i in 4:6 loop
        A[i-3] := seed[i];
      end for;
      A := Src.MatVecModM( Src.A2p127,A,  Src.m2);
      for i in 4:6 loop
        outseed[i] := A[i-3];
      end for;
      Src.SaveSeed(outseed);
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
CreateStream
</font></h3>
<p>
Initializes an RngStream g
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>CreateStream</b>()</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Reads the value of the package seed, initializes the RngStream and updates the package seed. 
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
g := CMRG.CreateStream();
</pre>
 
</HTML>
"));
    end CreateStream;

    function ResetStartStream "ResetStartStream"
      input RngStream g "RngStream";
      output RngStream gout "Updated RngStream";
    protected
      Integer i;
    algorithm
      gout := g;
      for i in 1:6 loop
        gout.Cg[i] := gout.Ig[i];
        gout.Bg[i] := gout.Ig[i];
      end for;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
ResetStartStream
</font></h3>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>ResetStartStream</b>(g)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Reinitializes the state of the RngStream g to the initial state of the stream.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
    for i in 1:3 loop
     (u[i],g) := CMRG.RandU01(g);
    end for;
    g := CMRG.ResetNextSubstream(g);
    for i in 4:6 loop
     (u[i],g) := CMRG.RandU01(g);
    end for;
    g := CMRG.ResetStartStream(g);
    for i in 7:9 loop
     (u[i],g) := CMRG.RandU01(g);
    end for;
 
  // u = {0.463941,0.733693,0.00275546,
          0.196378,0.218283,0.34092,
          0.463941,0.733693,0.00275546}
</pre>
 
</HTML>
"));
    end ResetStartStream;
    /*-------------------------------------------------------------------------*/
    function ResetNextSubstream "ResetNextSubstream"
      input RngStream g "RngStream";
      output RngStream gout "Updated RngStream";
    protected
      Integer i;
      Real[3] A;
    algorithm
      gout := g;
      for i in 1:3 loop
        A[i] := gout.Bg[i];
      end for;
      A := Src.MatVecModM( Src.A1p76, A, Src.m1);
      for i in 1:3 loop
        gout.Bg[i] := A[i];
      end for;
      for i in 4:6 loop
        A[i-3] := gout.Bg[i];
      end for;
      A := Src.MatVecModM( Src.A2p76,A,  Src.m2);
      for i in 4:6 loop
        gout.Bg[i] := A[i-3];
      end for;
      for i in 1:6 loop
        gout.Cg[i] := gout.Bg[i];
      end for;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
ResetNextSubstream
</font></h3>
 
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>ResetNextSubstream</b>(g)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
 Reinitializes the state of RngStream g to the beginning of the next sub-stream.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
    g2 := g1;
    g2 := CMRG.ResetNextSubstream(g2);
    for i in 1:3 loop
      (u1[i],g1) := CMRG.RandU01(g1);
      (u2[i],g2) := CMRG.RandU01(g2);
    end for;
    g1 := CMRG.ResetNextSubstream(g1);
    for i in 4:6 loop
      (u1[i],g1) := CMRG.RandU01(g1);
      (u2[i],g2) := CMRG.RandU01(g2);
    end for;
 
  // u1 = {0.89111  ,0.941477,0.194141,
  //       0.0368941,0.606502,0.597065}
  // u2 = {0.0368941,0.606502,0.597065,
  //       0.768019 ,0.628601,0.443735}
 
 </pre>
 
</HTML>
"));
    end ResetNextSubstream;
    /*-------------------------------------------------------------------------*/
    function ResetStartSubstream "ResetStartSubstream"
      input RngStream g "RngStream";
      output RngStream gout "Updated RngStream";
    protected
      Integer i;
    algorithm
      gout := g;
      for i in 1:6 loop
        gout.Cg[i] := gout.Bg[i];
      end for;
    annotation(preferredView="info",
      Documentation(info="<HTML>
 
<h3><font color=\"#008000\" size=5>
ResetStartSubstream
</font></h3>
 
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>ResetStartSubstream</b>(g)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
 Reinitializes the state of RngStream g to the beginning of the current sub-stream.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
    for i in 1:3 loop
      (u[i],g) := CMRG.RandU01(g);
    end for;
    g := CMRG.ResetStartSubstream(g);
    for i in 4:6 loop
      (u[i],g) := CMRG.RandU01(g);
    end for;
 
  // u = {0.584045,0.18949,0.698993,
  //      0.584045,0.18949,0.698993}
 
</pre>
 
 
</HTML>
"));
    end ResetStartSubstream;
    /*-------------------------------------------------------------------------*/
    function SetPackageSeed "SetPackageSeed"
      input Src.Seed seed "New package seed";
      output Integer out "result (0 == success)";
    protected
      Integer i;
    algorithm
      if (Src.CheckSeed( seed) <> 0) then
        out := -1;                    /* FAILURE */
      else
        Src.SaveSeed(seed);
        out := 0;
      end if;                       /* SUCCESS */
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
SetPackageSeed
</font></h3>
 
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>SetPackageSeed</b>(seed)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
 Changes the current seed stored in the file \"CMRGSeed.txt\".
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
  SetPackageSeed({54321,54321,54321,54321,54321,54321});
</pre>
 
</HTML>
"));
    end SetPackageSeed;
    /*-------------------------------------------------------------------------*/
    function SetSeed "SetSeed"
      input RngStream g "RngStream";
      input Src.Seed seed "New seed";
      output RngStream gout "Updated RngStream";
      output Integer out "result (0 == success)";
    protected
      Integer i;
    algorithm
      gout := g;
      if (Src.CheckSeed( seed) <> 0) then
        out := -1;  /* FAILURE */
      else
        for i in 1:6 loop
          gout.Cg[i] := seed[i];
          gout.Bg[i] := seed[i];
          gout.Ig[i] := seed[i];
       end for;
        out := 0;      /* SUCCESS */
      end if;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
SetSeed
</font></h3>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>SetSeed</b>(g,seed)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
 Changes the initial seed of the RngStream g. This modificatioin also changes the current state of the RngStream.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
  g2 := g1;
  for i in 1:3 loop
    (u1[i],g1) := CMRG.RandU01(g1);
    (u2[i],g2) := CMRG.RandU01(g2);
  end for;
  g2 := CMRG.SetSeed(g2,{2345,2435,2345,2345,2345,2345});
  for i in 4:6 loop
    (u1[i],g1) := CMRG.RandU01(g1);
    (u2[i],g2) := CMRG.RandU01(g2);
  end for;
  
  // u1 = {0.586819,0.848015,0.494309,0.609745,0.698027,0.812434}
  // u2 = {0.586819,0.848015,0.494309,0.813362,0.517802,0.947542}
</pre>
 
</HTML>
"));
    end SetSeed;
    /*-------------------------------------------------------------------------*/
    function AdvanceState
      "Advances the state of the RngStream g by k values. If e>0 , then k=2^e+c; if e<0, then k=-2^-e+c; and if e=0,then k=c"
      input RngStream g;
      input Real e;
      input Real c;
      output RngStream gout;
    protected
      Real[3,3] B1;
      Real[3,3] C1;
      Real[3,3] B2;
      Real[3,3] C2;
      Real[3] A;
    algorithm
      gout := g;
      if (e > 0) then
        B1 := Src.MatTwoPowModM( Src.A1p0, Src.m1, e);
        B2 := Src.MatTwoPowModM( Src.A2p0, Src.m2, e);
      else
        if (e < 0) then
          B1  := Src.MatTwoPowModM( Src.InvA1, Src.m1, -e);
          B2  := Src.MatTwoPowModM( Src.InvA2, Src.m2, -e);
        end if;
      end if;
      if (c >= 0) then
        C1 := Src.MatPowModM( Src.A1p0, Src.m1, c);
        C2 := Src.MatPowModM( Src.A2p0, Src.m2, c);
      else
        C1 := Src.MatPowModM( Src.InvA1, Src.m1, -c);
        C2 := Src.MatPowModM( Src.InvA2, Src.m2, -c);
      end if;
      if (e <> 0) then
        C1 := Src.MatMatModM( B1, C1, Src.m1);
        C2 := Src.MatMatModM( B2, C2, Src.m2);
      end if;
     for i in 1:3 loop
        A[i] := gout.Cg[i];
      end for;
      A := Src.MatVecModM( C1, A, Src.m1);
      for i in 1:3 loop
        gout.Cg[i] := A[i];
      end for;
      for i in 4:6 loop
        A[i-3] := gout.Cg[i];
      end for;
      A := Src.MatVecModM( C2,A,  Src.m2);
      for i in 4:6 loop
        gout.Cg[i] := A[i-3];
      end for;
    end AdvanceState;
    /*-------------------------------------------------------------------------*/
    function GetState "GetState"
      input RngStream g "RngStream";
      output Src.Seed seed "Current state of RngStream";
    protected
      Integer i;
    algorithm
      for i in 1:6 loop
        seed[i] := g.Cg[i];
      end for;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
GetState
</font></h3>
 
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>GetState</b>(g)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
 Returns the current state of RngStream g.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
  CMRG.GetState(g) =
  {2.9888e+009,1.4434e+009,3.39191e+009,6.80411e+007,1.94413e+009,1.91104e+008}
</pre>
 
</HTML>
"));
    end GetState;
    /*-------------------------------------------------------------------------*/
    function IncreasedPrecis "IncreasedPrecis"
      input RngStream g "RngStream";
      input Integer incp "Increased precission switch";
      output RngStream gout "Updated RngStream";
    algorithm
      gout := g;
      gout.IncPrec := incp;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
IncreasedPrecis
</font></h3>
<p>
Increases the precision of the numbers generated from RngStream g
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>IncreasedPrecis</b>(g,incp)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
When called with incp > 0, switches the RngStream to increased precision mode and the generated number will be in increased precision. <br>
To switch the increased precision mode back it is necessary to call this function with incp == 0.<br><br>
Since Modelica only have one precision for Real numbers, the effect of this function is only to generate alternate random numbers from the ones in the stream.
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
    g2 := CMRG.IncreasedPrecis(g2,1);
    for i in 1:6 loop
      (u1[i],g1) := CMRG.RandU01(g1);
      //(u2[i],g2) := CMRG.RandU01(g2);
    end for;
    for i in 1:3 loop
       (u2[i],g2) := CMRG.RandU01(g2);
    end for;
 
// u1 = {0.736348,0.970465,0.776025,0.860987,0.803047,0.574409}
// u2 = {0.736348,         0.776025,         0.803047}
</pre>
 
</HTML>
"));
    end IncreasedPrecis;
    /*-------------------------------------------------------------------------*/
    function SetAntithetic "SetAntithetic"
      input RngStream g "RngStream";
      input Integer a "Antithetic switch";
      output RngStream gout "Updated RngStream";
    algorithm
      gout := g;
      gout.Anti := a;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
SetAntithetic
</font></h3>
<p>
Change to antithetic generation mode.
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>SetAntithetic</b>(g,a)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
With a &lt;> 0, the RngStream g will generate antithetic numbers (U-1 instead of U)
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
    g2 := g1;
    g2 := CMRG.SetAntithetic(g2,1);
    u1 := CMRG.RandU01(g1);
    u2 := CMRG.RandU01(g2);
 
// u1 = 0.37564
// u2 = 0.62436
 
</pre>
 
</HTML>
"));
    end SetAntithetic;
    /*-------------------------------------------------------------------------*/
    function RandU01 "RandU01"
      input RngStream g "RngStream";
      output Real out "Generated uniform random number";
      output RngStream gout "Updated RngStream";
    algorithm
      if (g.IncPrec <> 0) then
        (out,gout)  := Src.U01d(g);
      else
        (out,gout)  := Src.U01(g);
      end if;
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
RandU01
</font></h3>
<p>
Uniform random number generation function
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>RandU01</b>(g)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Returns a uniform random number in the interval (0,1), and updates the state of RngStream g by one step
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
model Example 
  CMRG.RngStream g;
  Real u[5];
algorithm 
  when initial() then
    g := CMRG.CreateStream();
  end when;
  when time &lt;= 0 then
    for i in 1:5 loop
      (u[i],g) := CMRG.RandU01(g);
    end for;
  end when;
end Example;
 
// u = {0.988831,0.760616,0.855857,0.546418,0.868702}
 
</pre>
 
</HTML>
"));
    end RandU01;
    /*-------------------------------------------------------------------------*/
    function RandInt "RandInt"
      input RngStream g;
      input Integer i;
      input Integer j;
      output Integer out;
      output RngStream gout;
    protected
      Real u;
    algorithm
      (u,gout) := RandU01(g);
      out := i+ integer((j-i +1.0) * u);
    annotation(preferredView="info",
      Documentation(info="<HTML>
<h3><font color=\"#008000\" size=5>
RandInt
</font></h3>
<p>
Discrete distribution random number generation function
</p>
 
<h3><font color=\"#008000\">Syntax</font></h3>
<blockquote><pre><b>RandInt</b>(g,i,j)</pre></blockquote>
 
 
<h3><font color=\"#008000\">Description</font></h3>
<P>
Returns a rundom number from the discrete uniform distribution over the integer (i,i+1,...,j), using RngStream g
</P>
 
 
<h3><font color=\"#008000\">Examples</font></h3>
<pre>
model Example 
  CMRG.RngStream g;
  Real u[5];
algorithm 
  when initial() then
    g := CMRG.CreateStream();
  end when;
  when time &lt;= 0 then
    for i in 1:5 loop
      (u[i],g) := CMRG.RandInt(g);
    end for;
  end when;
end Example;
 
// u = {8,8,7,9,2}
 
</pre>
 
</HTML>
"));
    end RandInt;
    annotation(preferredView="info",
    Documentation(info="<HTML>
<p>
<b>CMRG</b> package is the implementation in Modelica of the RNG algorithm proposed by L'Ecuyer et al.
(<a href=\"http://www.iro.umontreal.ca/~lecuyer/myftp/papers/streams00.pdf\">pdf</a>)<br>
</p>
 
<p> A description of the package structure and its use is given in the <a href=\"Modelica://DESLib.RandomLib.CMRG.UsersGuide\">Users Guide</a>
 
<p>
<dl>
<dt><b>Author:</b></dt>
<dd><a href=\"http://www.euclides.dia.uned.es/vsanz\"> Victorino Sanz </a><br>
    Dpto. Inform&aacute;tica y Autom&aacute;tica, UNED<br>
    Juan del Rosal, 16<br>
    28040, Madrid<br>
    Spain<br>
    email: <A HREF=\"mailto:vsanz@dia.uned.es\">vsanz@dia.uned.es</A><br></dd>
</dl>
</p>
 
<p>
<dt><b>Copying:</b></dt>
<p>
<p>
Licensed by Victorino Sanz under the Modelica License 2 <br>
Copyright 2009, Victorino Sanz.
</p>
<p>
<i>This Modelica package is <u>free</u> software and the use is completely at <u>your ouwn risk;</u> it can be redistributed and/or modified under the terms of the Modelica license 2, 
see the license conditions (including the disclaimer of warranty) <a href=\"Modelica://DESLib.ModelicaLicense2\">here</a> or at <a href=\"http://www.modelica.org/licenses/ModelicaLicense2\">http://www.modelica.org/licenses/ModelicaLicense2</a>
</p>
<br>
</HTML>
"));
  end CMRG;

  package Examples
    "Examples of random uniform numbers and random variates generation"
      model CMRGSimple "generates 10 random numbers from an RngStream"
        CMRG.RngStream g "RngStream";
        Real u[10] "vector of random numbers";
        Boolean init( start = false);
      algorithm
      // initilization of the RngStream
        when initial() and not init then
          g := CMRG.CreateStream();
          init := true;
        end when;
        when sample(0,1) then
          for i in 1:10 loop
            // generation of uniform random numbers.
            (u[i],g) := CMRG.RandU01(g);
            Modelica.Utilities.Streams.print(String(u[i]));
          end for;
        end when;
      annotation (experiment(StopTime=10));
      end CMRGSimple;

      model Mean
      "generates random variates (one per second) and calculates the mean of the generated samples"
        CMRG.RngStream g "RngStream";
        Real u "random number";
        Integer Obs(start = 0) "number of observations";
        Real Mean( start = 0) "mean";
      initial algorithm
      // initilization of the RngStream
        g := CMRG.CreateStream();
      algorithm
      when sample(0,1) then
        (u,g) := Variates.GenerateVariate(13,g,5,1); // generate Normal variates with mean 5 and std 1
        Obs := Obs +1;
        Mean := ((Obs/(Obs+1))*Mean)+(u/(Obs+1));
      end when;
        annotation (experiment(StopTime=10000));
      end Mean;

    model VariatesSimple
      "generates 5 random variates with Exponential(5) distribution"
      Variates.Generator g "RNG";
      Real u[5] "vector of random variates";
    initial algorithm
      // initilization of the RngStream
        g := Variates.initGenerator();

    algorithm
        when time > 1 then
        for i in 1:5 loop
          // generation of uniform random numbers.
          (u[i],g) := Variates.Continuous.Exponential(g,8);
          Modelica.Utilities.Streams.print(String(u[i]));
        end for;
      end when;
      annotation (experiment(StopTime=2));
    end VariatesSimple;

    model VariatesSimple2
      "generates a discrete variate of each distribution at each integration step"
      Variates.Generator g "RNG";
      Real ConstantVariate;
      Real Bernoulli;
      Real DiscreteUniform;
      Real Binomial;
      Real Geometric;
      Real NegativeBinomial;
      Real Poisson;
      Real Uniform;
      Real Exponential;
      Real Erlang;
      Real Gamma;
      Real Weibull;
      Real Normal;
      Real LogNormal;
      Real Beta;
      Real Johnson;
      Real Triangular;

    initial algorithm
      // initilization of the RngStream
        g := Variates.initGenerator();

    algorithm
      when not initial() then
      (ConstantVariate,g) :=Variates.GenerateVariate(
                1,
                g,
                0.4);
      (Bernoulli,g) :=Variates.GenerateVariate(
                2,
                g,
                0.4);
      (DiscreteUniform,g) :=Variates.GenerateVariate(
                3,
                g,
                0,
                2);
      (Binomial,g) :=Variates.GenerateVariate(
                4,
                g,
                4,
                0.3);
      (Geometric,g) :=Variates.GenerateVariate(
                5,
                g,
                0.5);
      (NegativeBinomial,g) :=Variates.GenerateVariate(
                6,
                g,
                6,
                0.2);
      (Poisson,g) :=Variates.GenerateVariate(
                7,
                g,
                4.5);
      (Uniform,g) :=Variates.GenerateVariate(
                8,
                g,
                2,
                4);
      (Exponential,g) :=Variates.GenerateVariate(
                9,
                g,
                0.4);
      (Erlang,g) :=Variates.GenerateVariate(
                10,
                g,
                0.5,
                4);
      (Gamma,g) :=Variates.GenerateVariate(
                11,
                g,
                0.4,
                0.7);
      (Weibull,g) :=Variates.GenerateVariate(
                12,
                g,
                0.4,
                0.8);
      (Normal,g) :=Variates.GenerateVariate(
                13,
                g,
                0,
                1);
      (LogNormal,g) :=Variates.GenerateVariate(
                14,
                g,
                0,
                1);
      (Beta,g) :=Variates.GenerateVariate(
                15,
                g,
                0.6,
                0.4);
      (Johnson,g) :=Variates.GenerateVariate(
                16,
                g,
                0.1,
                0.5,
                1,
                4);
      (Triangular,g) :=Variates.GenerateVariate(
                17,
                g,
                1,
                3,
                6);
      end when;
      annotation (experiment(StopTime=0.1, __Dymola_Algorithm="Dassl"));
    end VariatesSimple2;

  model Test "Pierre L'Ecuyer test program for RngStreams"
      import RandomLib.CMRG.*;
    RngStream g1;
    RngStream g2;
    RngStream g3;
    RngStream gar1;
    RngStream gar2;
    RngStream gar3;
    RngStream gar4;
    Real germe[6] = { 1, 1, 1, 1, 1, 1};
    Real sum;
    Real sum3;
    Integer sumi;
    Real u1;
    Real u2;
    Real u3;
    Integer i1;
    function WriteState
      input CMRG.RngStream g;
    algorithm
      Modelica.Utilities.Streams.print("The current state of the Rngstream");
      Modelica.Utilities.Streams.print("Cg = {"+String(g.Cg[1],significantDigits=14)+","+String(g.Cg[2],significantDigits=14)+","+String(g.Cg[3],significantDigits=14)+","+String(g.Cg[4],significantDigits=14)+","+String(g.Cg[5],significantDigits=14)+","+String(g.Cg[6],significantDigits=14)+"}\n");
    end WriteState;
  algorithm
    when time > 0 then
      g1 := CreateStream();
      g2 := CreateStream();
      g3 := CreateStream();
      Modelica.Utilities.Streams.print("Initial states of g1, g2, and g3:\n");
      WriteState(g1);
      WriteState(g2);
      WriteState(g3);
      (u2,g2) := RandU01(g2);
      (u3,g3) := RandU01(g3);
      sum :=u2 + u3;
      for i in 1:12345 loop
        (u2,g2) := RandU01(g2);
      end for;
      g1 := AdvanceState(g1, 5, 3);
      Modelica.Utilities.Streams.print("State of g1 after advancing by 2^5 + 3 = 35 steps:");
      WriteState(g1);
      (u1,g1) := RandU01(g1);
      Modelica.Utilities.Streams.print("RandU01 (g1) = "+String(u1)+"\n");
      g1 :=  ResetStartStream(g1);
      for i in 1:35 loop
        g1 := AdvanceState(g1, 0,1);
      end for;
      Modelica.Utilities.Streams.print("State of g1 after reset and advancing 35 times by 1:");
      WriteState( g1);
      (u1,g1) := RandU01(g1);
      Modelica.Utilities.Streams.print("RandU01 (g1) = "+String(u1)+"\n");
      g1 := ResetStartStream(g1);
      for i in 1:35 loop
        (i1,g1) := RandInt(g1,1,10);
        sumi := sumi + i1;
      end for;
      Modelica.Utilities.Streams.print("State of g1 after reset and 35 calls to RandInt (1, 10):");
      WriteState(g1);
      Modelica.Utilities.Streams.print("   sum of 35 integers in [1, 10] = "+String(sumi)+"\n");
      sum := sum + (sumi / 100.0);
      (u1,g1) := RandU01(g1);
      Modelica.Utilities.Streams.print( "RandU01 (g1) = "+String(u1)+"\n");
      sum3 :=0.0;
      g1 := ResetStartStream( g1);
      g1 := IncreasedPrecis( g1, 1);
      sumi :=0;
      for i in 1:17 loop
        (i1,g1) := RandInt(g1,1,10);
        sumi := sumi + i1;
      end for;
      Modelica.Utilities.Streams.print("State of g1 after reset, IncreasedPrecis (1) and 17 calls");
      Modelica.Utilities.Streams.print(" to RandInt (1, 10):");
      WriteState( g1);
      g1 := IncreasedPrecis( g1, 0);
      (i1,g1) := RandInt(g1,1,10);
      Modelica.Utilities.Streams.print( "State of g1 after IncreasedPrecis (0) and 1 call to RandInt");
      WriteState( g1);
      sum3 :=sumi/10.0;
      g1 := ResetStartStream( g1);
      g1 := IncreasedPrecis( g1, 1);
      for i in 1:17 loop
        (u1,g1) := RandU01(g1);
        sum3 := sum3 + u1;
      end for;
      Modelica.Utilities.Streams.print("State of g1 after reset, IncreasedPrecis (1) and 17 calls");
      Modelica.Utilities.Streams.print(" to RandU01:");
      WriteState( g1);
      g1 := IncreasedPrecis( g1, 0);
      (u1,g1) := RandU01(g1);
      Modelica.Utilities.Streams.print("State of g1 after IncreasedPrecis (0) and 1 call to RandU01");
      WriteState( g1);
      sum := sum + sum3 / 10.0;
      sum3 := 0.0;
      Modelica.Utilities.Streams.print( "Sum of first 100 output values from stream g3:");
      for i in 1:100 loop
        (u3,g3) := RandU01(g3);
        sum3 := sum3 + u3;
      end for;
      Modelica.Utilities.Streams.print( "   sum = "+String(sum3)+"\n");
      sum := sum + sum3 / 10.0;
      Modelica.Utilities.Streams.print( "\nReset stream g3 to its initial seed.");
      g3 := ResetStartSubstream(g3);
      Modelica.Utilities.Streams.print( "First 5 output values from stream g3:");
      for i in 1:5 loop
        (u3,g3) := RandU01(g3);
        Modelica.Utilities.Streams.print(String(u3));
      end for;
      (u3,g3) := RandU01(g3);
      sum :=sum + u3;
      Modelica.Utilities.Streams.print("\nReset stream g3 to the next Substream, 4 times.");
      for i in 1:4 loop
        g3 := ResetNextSubstream( g3);
      end for;
      Modelica.Utilities.Streams.print("First 5 output values from stream g3, fourth Substream:");
      for i in 1:5 loop
        (u3,g3) := RandU01(g3);
        Modelica.Utilities.Streams.print(String(u3));
      end for;
      (u3,g3) := RandU01(g3);
      sum := sum + u3;
      Modelica.Utilities.Streams.print("\nReset stream g2 to the beginning of Substream.");
      g2 := ResetStartSubstream( g2);
      Modelica.Utilities.Streams.print(" Sum of 100000 values from stream g2 with double precision:   ");
      g2 := IncreasedPrecis( g2, 1);
      sum3 :=0.0;
      for i in 1:100000 loop
        (u2,g2) := RandU01(g2);
        sum3 := sum3 + u2;
      end for;
      Modelica.Utilities.Streams.print(String(sum3));
      sum := sum + sum3 / 10000.0;
      g2 := IncreasedPrecis( g2, 0);
      g3 := SetAntithetic( g3, 1);
      Modelica.Utilities.Streams.print(" Sum of 100000 antithetic output values from stream g3:   ");
      sum3 :=0.0;
      for i in 1:100000 loop
        (u3,g3) := RandU01(g3);
        sum3 := sum3 + u3;
      end for;
      Modelica.Utilities.Streams.print(String(sum3));
       sum := sum + sum3 / 10000.0;
      Modelica.Utilities.Streams.print("SetPackageSeed to seed = { 1, 1, 1, 1, 1, 1 }\n");
      SetPackageSeed(germe);
      Modelica.Utilities.Streams.print("\nCreate an array of 4 named streams and ");
      Modelica.Utilities.Streams.print("write their full state\n");
      gar1 := CreateStream();
      gar2 := CreateStream();
      gar3 := CreateStream();
      gar4 := CreateStream();
      WriteState(gar1);
      WriteState(gar2);
      WriteState(gar3);
      WriteState(gar4);
      Modelica.Utilities.Streams.print("Jump stream Galois (gar3) by 2^127 steps backward");
      gar3 := AdvanceState( gar3, -127, 0);
      WriteState( gar3);
      gar3 := ResetNextSubstream(gar3);
      (u1,gar1) := RandU01(gar1);
      sum := sum + u1;
      (u1,gar2) := RandU01(gar2);
      sum := sum + u1;
      (u1,gar3) := RandU01(gar3);
      sum := sum + u1;
      (u1,gar4) := RandU01(gar4);
      sum := sum + u1;
      Modelica.Utilities.Streams.print("--------------------------------------");
      Modelica.Utilities.Streams.print("Final Sum = "+String(sum)+"\n");
    end when;
  end Test;

    model VariatesSimple3
      "generates a discrete variate of each distribution at each integration step"
      Variates.Generator g "RNG";
      Real ConstantVariate;
      Real Bernoulli;
      Real DiscreteUniform;
      Real Binomial;
      Real Geometric;
      Real NegativeBinomial;
      Real Poisson;
      Real Uniform;
      Real Exponential;
      Real Erlang;
      Real Gamma;
      Real Weibull;
      Real Normal;
      Real LogNormal;
      Real Beta;
      Real Johnson;
      Real Triangular;

    initial algorithm
      // initilization of the RngStream
        g := Variates.initGenerator();
    equation
     //
    algorithm
      when time > 0 then
    (Exponential,g) :=Variates.GenerateVariate(
          9,
          g,
          0.4);
      end when;
      when not initial() then
      (ConstantVariate,g) :=Variates.GenerateVariate(
          1,
          g,
          0.4);
            (Bernoulli,g) :=Variates.GenerateVariate(
        2,
        g,
        0.4);

      (DiscreteUniform,g) :=Variates.GenerateVariate(
          3,
          g,
          0,
          2);
      (Binomial,g) :=Variates.GenerateVariate(
          4,
          g,
          4,
          0.3);
      (Geometric,g) :=Variates.GenerateVariate(
          5,
          g,
          0.5);
      (NegativeBinomial,g) :=Variates.GenerateVariate(
          6,
          g,
          6,
          0.2);
      (Poisson,g) :=Variates.GenerateVariate(
          7,
          g,
          4.5);
      (Uniform,g) :=Variates.GenerateVariate(
          8,
          g,
          2,
          4);

      (Erlang,g) :=Variates.GenerateVariate(
          10,
          g,
          0.5,
          4);
      (Gamma,g) :=Variates.GenerateVariate(
          11,
          g,
          0.4,
          0.7);
      (Weibull,g) :=Variates.GenerateVariate(
          12,
          g,
          0.4,
          0.8);
      (Normal,g) :=Variates.GenerateVariate(
          13,
          g,
          0,
          1);
      (LogNormal,g) :=Variates.GenerateVariate(
          14,
          g,
          0,
          1);
      (Beta,g) :=Variates.GenerateVariate(
          15,
          g,
          0.6,
          0.4);
      (Johnson,g) :=Variates.GenerateVariate(
          16,
          g,
          0.1,
          0.5,
          1,
          4);
      (Triangular,g) :=Variates.GenerateVariate(
          17,
          g,
          1,
          3,
          6);
      end when;
      annotation (experiment(StopTime=0.1, __Dymola_Algorithm="Dassl"));
    end VariatesSimple3;
  annotation(preferredView="info",
    Documentation(info="<HTML>
<p>
This package contains some examples that show the use of the RandomLib library and its components.
</p>
 
 
<p>
<em>CMRGSimple</em>, <em>Mean</em>, <em>VariatesSimple</em> and <em>VariatesSimple2</em> are very simple examples of uniform random number and random variates generation.
</P>
<p>
<em>Test</em> is a larger test case for uniform random number generation. This example is based in the test code provided by L'Ecuyer with his <a href=\"http://www.iro.umontreal.ca/~lecuyer/myftp/streams00/c/\">RngStream C implementation</a>.<br>
This example has been used for validating the implementation of the CMRG library in comparison with the results obtained with the mentioned C implementation.
</p>
 
 
<dl>
<dt><b>Author:</b></dt>
<dd><a href=\"http://www.euclides.dia.uned.es/vsanz\"> Victorino Sanz </a><br>
    Dpto. Informática y Automática, UNED<br>
    Juan del Rosal, 16<br>
    28040, Madrid<br>
    Spain<br>
    email: <A HREF=\"mailto:vsanz@dia.uned.es\">vsanz@dia.uned.es</A><br></dd>
</dl>
</p>
 
<p>
<dt><b>Copying:</b></dt>
<p>
<p>
Licensed by Victorino Sanz under the Modelica License 2 <br>
Copyright 2009, Victorino Sanz.
</p>
<p>
<i>This Modelica packages is <u>free</u> software and the use is completely at <u>your ouwn risk;</u> it can be redistributed and/or modified under the terms of the Modelica license 2, 
see the license conditions (including the disclaimer of warranty) <a href=\"Modelica://DESLib.ModelicaLicense2\">here</a> or at <a href=\"http://www.modelica.org/licenses/ModelicaLicense2\">http://www.modelica.org/licenses/ModelicaLicense2</a>
</p>
<br>
</HTML>
"));
  end Examples;
annotation(preferredView="info",
  Documentation(info="<HTML>
<p>
<b>RandomLib</b> is a Modelica package for uniform random number and random variates generation.
</p>
 
<p>
The structure of the package is:
<ul>
<li> The <a href=\"Modelica://DESLib.RandomLib.Variates\">Variates</a> package, for generating random variates.
<li> The <a href=\"Modelica://DESLib.RandomLib.CMRG\">CMRG</a> package, contains a uniform random number generator.
<li> The <a href=\"Modelica://DESLib.RandomLib.Examples\">Examples</a> package, with some examples and case studies.
</ul> 
 
<dl>
<dt><b>Author:</b></dt>
<dd><a href=\"http://www.euclides.dia.uned.es/vsanz\"> Victorino Sanz </a><br>
    Dpto. Inform&aacute;tica y Autom&aacute;tica, UNED<br>
    Juan del Rosal, 16<br>
    28040, Madrid<br>
    Spain<br>
    Email: <A HREF=\"mailto:vsanz@dia.uned.es\">vsanz@dia.uned.es</A><br>
    Website: <A HREF=\"http://www.euclides.dia.uned.es/vsanz\">www.euclides.dia.uned.es/vsanz</A></dd>
</dl>
</p>
 
 
<p>
<dt><b>Copying:</b></dt>
<p>
<p>
Licensed by Victorino Sanz under the Modelica License 2 <br>
Copyright 2009, Victorino Sanz.
</p>
<p>
<i>This Modelica package is <u>free</u> software and the use is completely at <u>your ouwn risk;</u> it can be redistributed and/or modified under the terms of the Modelica license 2, 
see the license conditions (including the disclaimer of warranty) <a href=\"Modelica://DESLib.ModelicaLicense2\">here</a> or at <a href=\"http://www.modelica.org/licenses/ModelicaLicense2\">http://www.modelica.org/licenses/ModelicaLicense2</a>
</p><br>
</HTML>
"),
    uses(Modelica(version="4.0.0")));
end RandomLib;
