/**********************************************************
* Author: 
*        Victorino Sanz 
*        Dpto. Informática y Automática, UNED
*        Juan del Rosal, 16
*        28040, Madrid
*        Spain
*        Email: vsanz@dia.uned.es
*
* Licensed by Victorino Sanz under the Modelica License 2
* Copyright © 2015, Victorino Sanz.
* 
***********************************************************/

#ifndef TLM
#define TLM

#include <CellularAutomataLib.c>


/******************************************************/
/* Transmission Line Model for Acoustic Propagation   */
/******************************************************/

//**************************************
// Cell state
// User may redefine the state of the cell as desired
typedef struct TLMState{
    int wall; // ==1 if cell is a wall
    double R; // reflection coefficient of the wall
    double in[4]; // cell inputs from N, E, S, W respectively
    double out[4]; // cell outputs from N, E, S, W respectively
}TLMState;



//**************************************
// CELLULAR SPACE FUNCTIONS

// Function that allocates memory and sets the default state for the cells
void * TLMDefaultState(){
    TLMState *s;
    s = (TLMState *)malloc(sizeof(TLMState));
    int i;
    
    //set the default values for state variables
    s->wall = 0; // not a wall
    s->R = 1;  // full reflection
    for(i=0;i<4;i++){
	s->in[i] = 0;  // no inputs
	s->out[i] = 0; // no outputs
    }
    return (void *)s;
}

int TLMDefault(void* space){
    void*  (*defaultState)() = &TLMDefaultState;
    CS2D_InitDefault(space, defaultState);
    return 1;   
}

// Function that describes the state of initialized cells (non-initialized cells will remain in the default state)
void* TLMInitialState(){
    TLMState *s;
    s = (TLMState *)malloc(sizeof(TLMState));
    int i;

    //set the initial values for state variables
    s->wall = 0; // not a wall
    s->R = 1;  // full reflection
    for(i=0;i<4;i++){
	s->in[i] = 0;
	s->out[i] = 10;
    }   
    //s->out[1] = 10; // initial wave propagation to the E
   
    return (void *)s;
}

// initializing function for the selected cell
void TLMInitial(void* space, int row, int col){
    void*  (*initialState)() = &TLMInitialState;
    ModelicaFormatMessage("INIT CELL [%d,%d]\n",row-1,col-1);
    CS2D_Init(space, row, col, initialState);
    return;
}


//***************************************************
// RULE FUNCTION
void *transition(int *change, void* cellstate, int row, int col, void** neighbors, int n_neighbors, void** inputs, int n_inputs, int *inputs_rcv){
    TLMState *out;
    TLMState *aux;
    TLMState *cs;
    TLMState **ne;
    double **in;
    int i,j,I,N;
    double u;
    double ZEROTHRESSHOLD = 0.0001;

    cs = (TLMState *)cellstate; // cell state
    ne = (TLMState **)neighbors; // neigbors cell state
    in = (double **)inputs;
    out = (TLMState *)malloc(sizeof(TLMState));
    aux = (TLMState *)malloc(sizeof(TLMState));
    
    ModelicaFormatMessage("STEP [%d,%d]\n",row,col);
    
    out->R = cs->R;
    out->wall = cs->wall;
    for(i=0;i<4;i++){
	out->in[i] = 0;
	out->out[i] = 0;
    }
    // PROPAGATION OF NEIGHBOR OUTPUTS AS NEW INPUTS
    if(ne[0] != NULL) out->in[0] = ne[0]->out[2]; else out->in[0] = 0;   
    if(ne[1] != NULL) out->in[1] = ne[1]->out[3]; else out->in[1] = 0;
    if(ne[2] != NULL) out->in[2] = ne[2]->out[0]; else out->in[2] = 0;
    if(ne[3] != NULL) out->in[3] = ne[3]->out[1]; else out->in[3] = 0;

 
    // PROPAGATION OF INPUTS TO OUTPUTS
    if (!out->wall){ // not a wall cell
	for(i=0;i<4;i++){
	    out->out[i] = 0;
	    for(j=0;j<4;j++){
		if (j != i)
		    out->out[i] += out->in[j]/2;
		else
		    out->out[i] -= out->in[j]/2;
	    }
	    if(out->out[i] < ZEROTHRESSHOLD) // set to 0 for small values to improve simulation
		out->out[i] = 0;
	}
    }else // cell is a wall
	for(i=0;i<4;i++)
	    out->out[i] = out->R*out->in[i]; // input reflected

	*change = 0;
	for(i=0;i<4;i++)
	    if(cs->out[i] != out->out[i])
	   *change = 1;

    return (void *)out;
}

// Function that returns the value to display in the graphical animation
double Display(void *cell){
    TLMState *s;
    double out;
    int i;

    s = (TLMState *)cell;
    // set out to the desired value calculated from the state variables
    if (cell == NULL)
	return 0;
    else{
	out = 0;
	for(i=0;i<4;i++)
	    out += s->out[i];
	return out;
    }
}


int TLMStep(void* space){
    void* (*rule)(int*,void*,int,int,void**,int,void**,int,int*) = &transition;
    double (*display)(void*) = &Display;
    
    return CS2D_Step(space,rule,display);
    //return 1;
}


/*********************************************/

//**************************************
// INTERFACE FUNCTIONS

// function to convert the state of a cell into a double value
double Output(void* space, int row, int col){
    double out;
    TLMState* s;
    
    s = (TLMState *)CS2D_GetState(space,row,col);
    
    out = 0;
    // include code
    return out;
}


// function to set an external value as an input to a cell
void VFull(void* space,int row, int col, double value,int input_id){
    double* inp;

    inp = (double *)malloc(sizeof(double));
    //convert the external value to values of the state variables
    *inp = value;
    CS2D_SetInput(space,row,col,(void*)inp,input_id);

    return;
}

void TLMInitS1(void* space, int row, int col, double value){
    TLMState* s;
    int pos;

    s = CS2D_GetState(space,row,col);
    // include code
    CS2D_SetState(space,row,col,s);
    return;
}


#endif
