/**********************************************************
* Author: 
*        Victorino Sanz 
*        Dpto. Informática y Automática, UNED
*        Juan del Rosal, 16
*        28040, Madrid
*        Spain
*        Email: vsanz@dia.uned.es
*
* Licensed by Victorino Sanz under the Modelica License 2
* Copyright © 2013, Victorino Sanz.
* 
***********************************************************/

#ifndef EPIDEMICS
#define EPIDEMICS

#include <CellularAutomataLib.c>


/****************************************/
/* EPIDEMICS CELLULAR AUTOMATA       */
/****************************************/

//**************************************
// Cell state
// User may redefine the state of the cell as desired
typedef struct EState{
    double N; //total population
    double S; // susceptible
    double I; // infected
    double R; // recovered
}EState;



//**************************************
// CELLULAR SPACE FUNCTIONS

// Function that allocates memory and sets the default state for the cells
void * DefaultState(){
    EState *s;
    s = (EState *)malloc(sizeof(EState));
    
    //set the default values for state variables
    s->N = 1;
    s->S = 1;
    s->I = 0;
    s->R  = 0;
    return (void *)s;
}

int EDefault(void* space){
    void*  (*defaultState)() = &DefaultState;
    CS2D_InitDefault(space, defaultState);
    return 1;
    
}

// Function that describes the state of initialized cells (non-initialized cells will remain in the default state)
void* InitialState(){
    EState *s;
    s = (EState *)malloc(sizeof(EState));
    
    // set the values of state variables for initialized cells
    s->N = 1;
    s->S = 0.7;
    s->I = 0.3;
    s->R  = 0;

    return (void *)s;
}

// initializing function for the selected cell
void Ini(void* space, int row, int col){
    void*  (*initialState)() = &InitialState;
    CS2D_Init(space, row, col, initialState);
    return;
}

//***************************************************
// RULE FUNCTION
void *transition(int *change, void* cellstate, int row, int col, void** neighbors, int n_neighbors, void** inputs, int n_inputs, int *inputs_rcv){
    EState *out;
    EState *cs;
    EState **ne;
    EState **in;
    double epsilon = 0.4;
    double v = 0.6;
    double c = 1;
    double m = 0.5;
    double mu  = c*m*v;
    double sum;
    int i;

    cs = (EState *)cellstate;
    ne = (EState **)neighbors;
    in = (EState **)inputs;
    out = (EState *)malloc(sizeof(EState));
    
    // define the behavior of the model during transitions
    sum = 0;
    for(i=0;i<n_neighbors;i++){
	if(neighbors[i] != NULL){
	    //sum = sum + (ne[i]->N/cs->N)*mu*ne[i]->I;
	    sum = sum + mu*ne[i]->I;
	}
    }

    out->I = (1-epsilon)*cs->I + v*cs->S*cs->I + cs->S*sum;
    out->S = cs->S - v*cs->S*cs->I -cs->S*sum;
    out->R = cs->R + epsilon*cs->I;

    *change = 1;

    return (void *)out;
}

// Function that returns the value to display in the graphical animation
double Display(void *cell){
    EState *s;
    double out;

    s = (EState *)cell;
    // set out to the desired value calculated from the state variables
    if (s == NULL)
	return 0;
    else
	return s->I;
}

int Step(void* space){
    void* (*rule)(int*,void*,int, int, void**,int,void**,int,int*) = &transition;
    double (*display)(void*) = &Display;
    
    CS2D_Step(space,rule,display);
    return 1;
}

/*********************************************/

//**************************************
// INTERFACE FUNCTIONS


// function initialize a cell using an external value
void ExtInit(void* space,int row, int col, double value){
    EState* s;

    s = (EState *)CS2D_GetState(space,row,col);
    // assign external value to the state variables
    s->I = value;
    CS2D_Init(space,row,col,NULL);
    return;
}


// Function to set a new input to a cell, from other cellular space
void SetInput(void* Fspace, int Frow, int Fcol, void* Tspace, int Trow, int Tcol, int input_id){
    EState* s;
    EState* inp;
    
    inp = (EState *)malloc(sizeof(EState));
    s = (EState *)CS2D_GetState(Fspace,Frow,Fcol);
    *inp = *s;
    CS2D_SetInput(Tspace,Trow,Tcol,(void*)inp,input_id);
}


// function to set an external value as an input to a cell
void ExtInput(void* space,int row, int col, double value,int input_id){
    EState* inp;

    inp = (EState *)malloc(sizeof(EState));
    //convert the external value to values of the state variables
    inp->I = value;

    CS2D_SetInput(space,row,col,(void*)inp,input_id);

    return;
}

// function to convert the state of a cell into a double value
double OutputS(void* space, int row, int col){
    double out;
    EState* s;
    
    s = (EState *)CS2D_GetState(space,row,col);
    
    out = s->S;
    
    return out;
}

// function to convert the state of a cell into a double value
double OutputI(void* space, int row, int col){
    double out;
    EState* s;
    
    s = (EState *)CS2D_GetState(space,row,col);
    
    out = s->I;
    
    return out;
}

// function to convert the state of a cell into a double value
double OutputR(void* space, int row, int col){
    double out;
    EState* s;
    
    s = (EState *)CS2D_GetState(space,row,col);
    
    out = s->R;
    
    return out;
}

#endif
