/**********************************************************
* Author: 
*        Victorino Sanz 
*        Dpto. Informática y Automática, UNED
*        Juan del Rosal, 16
*        28040, Madrid
*        Spain
*        Email: vsanz@dia.uned.es
*
* Licensed by Victorino Sanz under the Modelica License 2
* Copyright © 2013, Victorino Sanz.
* 
***********************************************************/

#ifndef FOREST
#define FOREST

#include <CellularAutomataLib.c>
#include <Behave.c>

/****************************************/
/* FOREST CELLULAR AUTOMATA       */
/****************************************/

//**************************************
// Cell state
// User may redefine the state of the cell as desired
typedef struct State{
    Behave *b;
    int s; //  cell state (0 unburned, 1 burning, 2 burned)
    int wet; // 0 not wet, 1 wet
    double windspeed;
    int winddir;
    double slope;
    int timetoWet; // time to wet
    int *timetoSpread; // number of steps to reach neighbors
    int fuelmodel;
}State;



//**************************************
// CELLULAR SPACE FUNCTIONS

// Function that allocates memory and sets the default state for the cells
void * DefaultState(){
    State *s;
    s = (State *)malloc(sizeof(State));
    
    //set the default values for state variables
    s->s = 0; // unburned
    s->wet = 0; // not wet
    s->windspeed = 8;//2.2352; 
    s->winddir = 180; // blows FROM south
    s->slope = 0;
    s->timetoWet = 9999; //
    s->timetoSpread = NULL;
    s->fuelmodel = 4;

    s->b = (Behave *)malloc(sizeof(Behave));
    initialValues(s->b);
    setFuelModel(s->fuelmodel,s->b);
    setWindSpeed(s->windspeed,s->b);
    setWindDir(s->winddir,s->b);
    setSlope(s->slope,s->b);
    calcRothermel(s->b);     
    return (void *)s;
}

int FDefault (void * space){
    void*  (*defaultState)() = &DefaultState;
    //    CS2D_SetDefault(space, defaultState);
    CS2D_InitDefault(space, defaultState);
    return 1;
}

// Function that describes the state of initialized cells (non-initialized cells will remain in the default state)
void * InitialState(){
    State *s;
    
    s = DefaultState();
        // set the values of state variables for initialized cells
    s->s = 1; // cell burning
    
    return (void *)s;
}

// initializing function for the selected cell
void Ini(int space, int row, int col){
    void*  (*initialState)() = &InitialState;
    void*  (*defaultState)() = &DefaultState;
    int i,j;
    
    //for(i=1;i<=row;i++)
    //	for(j=1;j<=col;j++){
    //	    if(i == (row-floor(row/6)) && j == (col/2))
		CS2D_Init(space, row, col, initialState);
		//	    else
		//CS2D_Init(space, i, j, defaultState);
		//	}
    return;
}

//***************************************************
// RULE FUNCTION
void *transition(int *change, void* cellstate, int row, int col, void** neighbors, int n_neighbors, void** inputs, int n_inputs, int *inputs_rcv){
    State *out;
    State *cs;
    State **ne;
    double **in;
    int ne_burning; // num of neighbors burning
    int ne_n; // num of neigbors
    double R[8];
    double D = 15; // distance between neighbors
    int i;

    cs = (State *)cellstate;
    ne = (State **)neighbors;
    in = (double **)inputs;
    out = (State *)malloc(sizeof(State));
    
    // define the behavior of the model during transitions
    
    *out = *cs;

    //calcRothermel(out->b); 
    if(n_inputs > 0){
	// INPUT[2] and INPUT[3] wind direction and speed changed
	if (inputs_rcv[1]){ // direction
	    out->winddir = *in[1]; //	
	}
	if (inputs_rcv[2]){ // speed
	    out->windspeed = *in[2]; // 
	}
	
	// INPUT[1] water poured or firefighting action
	if (inputs_rcv[0]){
	    out->timetoWet = *in[0]; // reset time to wet value
	    if (out->s == 0) // if unburned, set to wet
		out->wet = 1;
	}
	if (out->timetoWet <= 0)
	    out->wet = 1;    
	out->timetoWet--;

	// re-calculate Rothermel model 
	setWindSpeed(out->windspeed,out->b);
	setWindDir(out->winddir,out->b);
	calcRothermel(out->b);    	
    }

    // ACTIONS
    if (out->s == 0 && !out->wet){ // unburned and not wet
	// INPUT[3] External ignition
	/*	if(n_inputs > 0)
	    if (inputs_rcv[3]){
		out->s = 1;
		}*/
	// Spread from neighbor
	for(i=0;i<n_neighbors;i++){
	    if(neighbors[i] != NULL){
		if(ne[i]->timetoSpread != NULL)
		    if(ne[i]->timetoSpread[mod(i+4,n_neighbors)] == 0 && !out->wet){ // not wet and fire spread from neighbor
			out->s = 1; //starts burning
		    }
	    }
	}
    }else if (cs->s == 1){ // burning
	if(out->timetoSpread == NULL){ // just started burning
	    out->timetoSpread = (int *)malloc(sizeof(int)*n_neighbors);
	    for(i=0;i<n_neighbors;i++){
		R[i] = getRosInDir(out->b->ros,out->b->sdr,-45+(i*45),out->b->ecc);
		if (i % 2){
		    out->timetoSpread[i] = (D/R[i]); // linear connection between cells
		}else{
		    out->timetoSpread[i] = (sqrt(2*pow(D,2))/R[i]); // diagonal connection between cells
		}	
	    }
	}else{ // continue burning
	    for(i=0;i<n_neighbors;i++){
		out->timetoSpread[i]--;
	    }	
	}
	if(!out->wet){
	    ne_burning = 0; 
	    for(i=0;i<n_neighbors;i++){
		if (ne[i]->s > 0 || ne[i]->wet) // neighbor burning, burned or wet
		//if ((out->timetoSpread[i] <= 0) || (ne[i]->wet))
		    ne_burning++;
	    }
	    if(ne_burning == n_neighbors) //
		out->s = 2; // to burned because all neighbors are burning, burned or wet
	}else{
	    out->s = 2; // to burned due to wet
	}
    }else if (out->s == 2){ // burned
    }
    
    *change = 1;
    if (out->s == 2 || out->s == 0)
	*change = 0;
    
    return (void *)out;
}

// Function that returns the value to display in the graphical animation
double Display(void *cell){
    State *s;
    double out;

    s = (State *)cell;
    // set out to the desired value calculated from the state variables
    if (cell == NULL)
	out = 0;
    else{
	out = s->s;
	if (s->wet) 
	    out = out+0.5;
    }
    return out;
}

int Step(void* space){
    void* (*rule)(int *, void*,int,int,void**,int,void**,int,int*) = &transition;
    double (*display)(void*) = &Display;
    
    return CS2D_Step(space,rule,display);
    //return 1;
}

/*********************************************/

//**************************************
// INTERFACE FUNCTIONS


// function initialize a cell using an external value
void ExtInit(void* space,int row, int col, double value){
    State* s;

    s = (State *)CS2D_GetState(space,row,col);
    // assign external value to the state variables
    s->s = value;
    CS2D_SetState(space,row,col,s);
    return;
}


// Function to set a new input to a cell, from other cellular space
void SetInput(void* Fspace, int Frow, int Fcol, void* Tspace, int Trow, int Tcol, int input_id){
    State* s;
    State* inp;
    
    inp = (State *)malloc(sizeof(State));
    s = (State *)CS2D_GetState(Fspace,Frow,Fcol);
    *inp = *s;
    CS2D_SetInput(Tspace,Trow,Tcol,(void*)inp,input_id);
}


// function to set an external value as an input to a cell
void ExtInput(void* space,int row, int col, double value,int input_id){
    double* inp;

    inp = (double *)malloc(sizeof(double));
    //convert the external value to values of the state variables
    *inp = value;

    CS2D_SetInput(space,row,col,(void*)inp,input_id);

    return;
}

// function to convert the state of a cell into a double value
double Output(void* space, int row, int col){
    double out;
    
    out = Display(CS2D_GetState(space,row,col));
    
    if (out > 0.5) // cell may be wet (out = 0.5)
	return 1; // cell burning or burned
    else
	return 0; 
}

#endif
