#ifndef EXTPDEVS
#define EXTPDEVS

#include "pdevs_ports.c";

typedef struct{
    port *IN; // input ports
    port *OUT; // output ports
    int nin,nout; // number of input and output ports

    int nstate; // number of state variables
    double *state; // state variables
    double sigma; // interval to next internal event
    
    double* (*intT)(double*,int); // (state, nstate) internal transition function
    double* (*extT)(double*,int,double, port *,int); // (state, nsatate, elapsedtime, IN, nin) external transition function
    double* (*conT)(double*,int,double, port *,int); // (state, nsatate, elapsedtime, IN, nin) confluent transitio nfunction
    void (*out)(double*,int,port *,int);//,void**); // (state, nstate, OUT, nout) output function
    double (*timeadvance)(double*,int); // (state, nstate) time advance function 
    
}atomic_model;


typedef struct{
    double t; // time for the event
    atomic_model *m; // model involved in the event
    int f; // function to execute (0 intT, 1 extT, 2 conT, 3 out, 4 timeadvance)
    struct calendar_event *prev;
    struct calendar_event *next;
}calendar_event;

typedef struct{
    double T; // current simulated time
    double Tend; // end time for the simulation
    double Tnext; // time for next transition in the calendar
    double Tlast; // time of last transition performed

    int nM;
    atomic_model **M; // list of references to atomic models 
    
    // event calendars
    int nint;
    calendar_event *INT; // internal transitions
    int next;
    calendar_event *EXT; // external transitions
    int sizecon;
    calendar_event *CON; // confluent transitions
    int nrest;
    calendar_event *REST; // calendar for the events in the simulation
    
    
}pdevs_simulator;

/***************************/

int atomic_init(atomic_model *m, int nin, int nout, int nstate, double *initstate){
    int i;
    
    m = (m*)malloc(sizeof(atomic_mode));
    m->IN = (port*)malloc(nin*sizeof(port));
    m->nin = nin;
    for(i=0; i<nin;i++){
	m->IN[i] = initPort(m);
    }
    m->OUT = (port*)malloc(nout*sizeof(port));
    m->nout = nout;
    for(i=0; i<nout;i++){
	m->OUT[i] = initPort(m);
    }
    m->nstate = nstate;
    m->state = (double *)malloc(nstate*sizeof(double));
    for(i=0;i<nstate;i++){
	m->state[i] = initstate[i];
    }
    return 0;
}


void sim_addmodel(pdevs_simulation * s, atomic_model *m){

    s->nM++;
    s->M = (atomic_model**)realloc(s->M,s->nM*sizeof(atomic_model*));
    s->M[s->nM-1] = m;
}

pdevs_simulator * sim_init(int tend){
    pdevs_simulator *s;
    int i;
    double ta;
    calendar_event *e;
    
    s = (pdevs_simulator*)malloc(sizeof(pdevs_simulator));
    s->T = 0;
    s->Tend = tend;
    s->INT = NULL;
    s->nint = 0;
    s->EXT = NULL;
    s->next = 0;
    s->CON = NULL;
    s->ncon = 0;

    // schedule new internal events for all models
    for(i=0;i<s->nM;i++){
	ta = s->M[i]->timeadvance(s->M[i]->state,s->M[i]->nstate);
	e = (calendar_event*)malloc(sizeof(calendar_event));
	e->t = s->T+ta;
	e->m = s->M[i];
	e->f = 0;
	calendar_insert(s->INT,e); // schedule new internal event in INT
    }
    s->Tnext = s->INT->t; // update time for next internal event
    // run transitions at T==0
    while (s->Tnext <= s->T){
	s->Tlast = s->T;
	s->Tnext = sim_step(s);
    }
    return s;
}

// simulation step (returns the time for the next transition)
double sim_step(pdevs_simulator* s){
    int i,j;
    int ndesg;
    void **Ndest;

    s->Tlast = s-T;
    s->T = s->Tnext; // advance simulation time
    while (s->Tnext <= s->T){
	sim_run_outputs(s);
	sim_run_transitions(s);
	s->Tnext = s->INT->t;
    }
    
    return s->Tnext;
}

int sim_run_outputs(pdevs_simulator *s){
    int i,p;
    calendar_event *aux;
    int ndest;
    void **Ndest = NULL;
    atomic_model *m;
    calendar_event *ev;
    
    aux = s->INT;
    // evaluate output function for all models in INT
    for(i=0; i<s->nint; i++){
	m = aux->m;
	m->out(m->state,m->nstate,m->OUT,m->nout); // generate outputs
	for(p=0;p<m->nout;p++){
	    ndest += send(m->OUT[p],Ndest); // send outputs and store destinations
	}
	aux = aux->next;
    }
    //  update EXT and CON, based on destinations that received inputs
    for(i=0; i<ndest; i++){
	m = (atomic_model*)Ndest[i];
	if(calendar_isinT(s->INT,m,s->T)){
	    // already in INT, move to CON
	    ev = calendar_remove(s->INT,m);
	    ev->f = 2;
	    calendar_insert(s->CON,ev);
	}else{
	    // not in INT, insert in EXT
	    ev = (calendar_event*)malloc(sizeof(calendar_event*));
	    ev->T = s->T;
	    ev->m = m;
	    ev->f = 1;
	    calendar_insert(s->EXT,ev);
	}
    }
    return 0;
}

int sim_run_transitions(pdevs_simulator *s){
    int i;
    calendar_event *ev;
    atomic_model *m;
    atomic_model **mtran;
    int nmtran;
    double ta;

    nmtran = 0;
    mtran = (atomic_model*)malloc(s->nM*sizeof(atomic_model*));

    // INT
    ev = calendar_removefirstT(s->INT,s->T);
    while(ev != NULL){
	m = ev->m;
	mtran[nmtran++] = m;
	m->intT(m->state,m->nstate);
	ev = calendar_removefirstT(s->INT,s->T);
    }
    // EXT
    for(i=0; i<s->next; i++){
	ev = calendar_removefirst(s->EXT);
	m = ev->m;
	mtran[nmtran++] = m;
	m->extT(m->state,m->nstate,s->T-s->Tlast,m->IN,m->nin);
    }
    // CON
    for(i=0; i<s->ncon; i++){
	ev = calendar_removefirst(s->CON);
	m = ev->m;
	mtran[nmtran++] = m;
	m->conT(m->state,m->nstate,s->T-s->Tlast,m->IN,m->nin);
    }
    // schedule new internal events for transitioned models
    for(i=0;i<s->nmtran;i++){
	ta = mtran[i]->timeadvance(mtran[i]->state,mtran[i]->nstate);
	e = (calendar_event*)malloc(sizeof(calendar_event));
	e->t = s->T+ta;
	e->m = mtran[i];
	e->f = 0;
	calendar_insert(s->INT,e); // schedule new internal event in INT
    }
    free(mtran);
    return 0;
}


int calendar_isin(calendar_event* cal, atomic_model *m){
    int i = 0;
    calenar_event *aux;
    int found = 0;
    
    aux = cal;
    while(aux != NULL){
	if(aux->m == m)
	    return 1;
	else
	    aux = aux->next;
    }
    return 0;
   
}

int calendar_isinT(calendar_event* cal, atomic_model *m, double T){
    int i = 0;
    calenar_event *aux;
    int found = 0;
    
    aux = cal;
    while(aux != NULL && aux->t <= T){
	if(aux->m == m)
	    return 1;
	else
	    aux = aux->next;
    }
    return 0;
   
}

int calendar_insert(calendar_event* cal, calendar_event* ev){
    calendar_event * aux;
    int i;
    
    aux = cal;
    while(aux != NULL){
	if (aux->t > ev->t){ // insert before aux
	    if (aux->prev == NULL) // first event in cal
		cal = ev;
	    ev->next = aux;
	    ev->prev = aux->prev;
	    aux->prev->next = ev;
	    aux->prev = ev;
	    break;
	}else if (aux->next == NULL){ //insert at the end
	    ev->next = aux->next;
	    ev->prev = aux;
	    aux->next = ev;
	    break;
	}else
	    aux = aux->next;
    }
    return 0;
}
 
calendar_event * calendar_removefirst(calendar_event* cal){
    calendar_event * aux;
    
    aux = cal;
    cal = aux->next;
    aux->next->prev = NULL;
    aux->next = NULL;
    aux->prev = NULL;
    
    return aux;
}

calendar_event * calendar_removefirstT(calendar_event* cal, double T){
    calendar_event * aux;
    
    aux = cal;
    if (aux->T <= T){
	cal = aux->next;
	aux->next->prev = NULL;
	aux->next = NULL;
	aux->prev = NULL;
    }else
	aux = NULL;
    
    return aux;
}

 calendar_event * calendar_remove(calendar_event* cal, atomic_model *m){
     calendar_event * aux;

     aux = cal;
     while(aux != NULL){
	 if(aux->m == m){
	     if(aux->prev == NULL){//first element, update cal
		 cal = aux->next;
		 aux->next->prev = NULL;
	     }else{
		 aux->prev->next = aux->next;
		 if(aux-next != NULL) // not the last element
		     aux->next->prev = aux->prev;
	     }
	     aux->next = NULL;
	     aux->prev = NULL;
	     break;
	 }else
	     aux = aux->next;
     }
     return aux;
}



#endif










